from functools import wraps
import numpy as np
from scipy.spatial import cKDTree
from scipy.spatial.distance import cdist
from numba import njit
import MDAnalysis as mda
import chilife


def clash_only(func):
    """Decorator to convert a Lennard-Jones style clash evaluation function into a chiLife compatible energy func.

    Parameters
    ----------
    func : callable
        Python function object that computes the (modified) Lennard-Jones potential given arrays of atom pair
        distances, `r`, rmin values `rmin`, energies, `eps`, and additional keyword arguments.

    Returns
    -------
    energy_func : callable
        The original input function now wrapped to accept a protein object and, optionally, a RotamerEnsemble/SpinLabel
        object.
    """

    @wraps(func)
    def energy_func(system, **kwargs):
        """
        Parameters
        ----------
        system : MDAnalysis.Universe, MDAnalysis.AtomGroup, RotamerEnsemble, Protein
            Molecular system for which you wish to calculate the energy of

        **kwargs : dict
            Additional arguments to pass to the wrapped function

        Returns
        -------
        E : numpy.ndarray
            Array of energy values (kcal/mol) for each rotamer in an ensemble or for the whole system.
        """
        internal = kwargs.pop("internal", False)
        rmax = kwargs.get("rmax", 10)
        forgive = kwargs.get("forgive", 1)
        _protein = kwargs.pop('protein', None)

        if isinstance(system, (chilife.RotamerEnsemble, chilife.dRotamerEnsemble)):
            if system.protein is None and _protein is not None:
                system.protein = _protein

            r, rmin, eps = prep_external_clash(system)
            E = func(r, rmin, eps, **kwargs)

            if internal:
                r, rmin, eps, shape = prep_internal_clash(system)
                Einternal = func(r, rmin, eps, **kwargs)
                E = np.concatenate((E, Einternal), axis=1)

            system.atom_energies = E.reshape(len(E), len(system.side_chain_idx), -1).sum(axis=2)
            E = E.sum(axis=1)


        elif isinstance(system, (mda.Universe, mda.AtomGroup, chilife.Protein)):
            bonds = {(a, b) for a, b in system.atoms.bonds.indices}
            tree = cKDTree(system.atoms.positions)
            pairs = tree.query_pairs(rmax)
            pairs = pairs - bonds
            pairs = np.array(list(pairs))

            r = np.linalg.norm(
                system.atoms.positions[pairs[:, 0]]
                - system.atoms.positions[pairs[:, 1]],
                axis=1,
                )

            lj_radii_1 = chilife.get_lj_rmin(system.atoms.types[pairs[:, 0]])
            lj_radii_2 = chilife.get_lj_rmin(system.atoms.types[pairs[:, 1]])

            lj_eps_1 = chilife.get_lj_eps(system.atoms.types[pairs[:, 0]])
            lj_eps_2 = chilife.get_lj_eps(system.atoms.types[pairs[:, 1]])

            join_rmin = chilife.get_lj_rmin("join_protocol")[()]
            join_eps = chilife.get_lj_eps("join_protocol")[()]

            rmin_ij = join_rmin(lj_radii_1 * forgive, lj_radii_2 * forgive, flat=True)
            eps_ij = join_eps(lj_eps_1, lj_eps_2, flat=True)
            E = func(r, rmin_ij, eps_ij, **kwargs)
            E = E.sum()

        else:
            raise TypeError(f'Energy evaluations of a {type(system)} object are not supported at this time. Please '
                            f'pass an chilife.RotamerEnsemble, mda.Universe, mda.AtomGroup or chilife.Protein')

        return E

    return energy_func


@clash_only
@njit(cache=True)
def get_lj_energy(r, rmin, eps, forgive=1, cap=10, rmax=10):
    """Return a vector with the energy values for the flat bottom lenard-jones potential from a set of atom pairs with
    distance `r`, rmin values of `rmin` and epsilon values of `eps`.

    Parameters
    ----------
    r : numpy.ndarray
        Vector of inter-atomic distances between non-bonded atoms of a system.
    rmin : numpy.ndarray
        Vector of rmin parameters, in angstoms,  corresponding to atoms pairs of r.
    eps : numpy.ndarray
        Vector of epsilon parameters corresponding to atom pairs of r
    forgive : numpy.ndarray
        The `forgive` factor is a softening term to mitigate rigid body artifacts. It is set to a value between 0 1nd 1
        and modifies the rmin parameter of all atom pairs in `r` to be the fraction, `forgive` of the thair original
        value. This allows atoms to be closer than otherwise allowed to prevent explosion of the jennard-jones repulsion
        in situations that would otherwise be resolved with minor atomic displacements.
    cap : flaot
        Maximum allowed energy factor. Sets a cap on the maximum energy contribution of one atom pair interaction as
        another mechanism for softening. `cap` is sets the actual max value as a multiple of the `eps` parameter. i.e.
        the maximum allowed energy from a single atom pair interaction is the `eps` parameter multiplied by `cap`.
    rmax : float
        Maximum distance to consider for potential calculation. Any atom pairs with r > rmax will be set to 0.

    Returns
    -------
    lj_energy: numpy.ndarray
        Vector of atom pair energies calculated using the modified lj potential function.

    """
    r = np.atleast_2d(r)
    lj_energy = np.zeros_like(r)
    rmin_lower = forgive * rmin

    # Piecewise function for flat lj potential near rmin
    for j, rotamer in enumerate(r):
        for i in range(len(rotamer)):
            if rotamer[i] < rmin_lower[i]:
                lj = rmin_lower[i] / rotamer[i]
                lj = lj * lj * lj
                lj = lj * lj
                lj_energy[j, i] = np.minimum(eps[i] * (lj**2 - 2 * lj), cap * eps[i])

            elif rmin_lower[i] <= rotamer[i] < rmin[i]:
                lj_energy[j, i] = -eps[i]
            elif rotamer[i] < rmax:
                lj = rmin[i] / rotamer[i]
                lj = lj * lj * lj
                lj = lj * lj
                lj_energy[j, i] = eps[i] * (lj**2 - 2 * lj)

    return lj_energy


@clash_only
@njit(cache=True)
def get_lj_scwrl(r, rmin, eps, forgive=1):
    """Calculate a scwrl-like lenard-jones potential from a set of atom pairs with distance `r`, rmin values of `rmin`
    and epsilon values of `eps`.

    Parameters
    ----------
    r : numpy.ndarray
        Vector of inter-atomic distances between non-bonded atoms of a system.
    rmin : numpy.ndarray
        Vector of rmin parameters, in angstoms,  corresponding to atoms pairs of r.
    eps : numpy.ndarray
        Vector of epsilon parameters corresponding to atom pairs of r
    forgive : numpy.ndarray
        The `forgive` factor is a softening term to mitigate rigid body artifacts. It is set to a value between 0 1nd 1
        and modifies the rmin parameter of all atom pairs in `r` to be the fraction, `forgive` of the thair original
        value. This allows atoms to be closer than otherwise allowed to prevent explosion of the jennard-jones repulsion
        in situations that would otherwise be resolved with minor atomic displacements.

    Returns
    -------
    lj_energy: numpy.ndarray
        Vector of atom pair energies calculated using the modified lj potential function.
    """
    r = np.atleast_2d(r)
    lj_energy = np.empty_like(r)
    rmin_lower = rmin * forgive
    for j, rotamer in enumerate(r):
        # Piecewise function for flat lj potential near rmin
        for i in range(len(rotamer)):
            rat = rotamer[i] / (rmin_lower[i] / 1.12246204831)
            if rat < 0.8254:
                lj_energy[j, i] = 10 * eps[i]
            elif rat <= 1:
                lj_energy[j, i] = 57.273 * (1 - rat) * eps[i]
            elif rat < 10 / 9:
                lj_energy[j, i] = eps[i] * (10 - 9 * rat) ** (57.273 / (9 * eps[i])) - eps[i]
            elif rat < 4 / 3:
                lj_energy[j, i] = (eps[i] / 4) * (9 * rat - 10) ** 2 - eps[i]
            else:
                lj_energy[j, i] = 0

    return lj_energy


@clash_only
@njit(cache=True)
def get_lj_rep(r, rmin, eps, forgive=0.9, cap=10):
    """Calculate only the repulsive terms of the lenard-jones potential from a set of atom pairs with distance `r`,
    rmin values of `rmin` and epsilon values of `eps`.

    Parameters
    ----------
    r : numpy.ndarray
        Vector of inter-atomic distances between non-bonded atoms of a system.
    rmin : numpy.ndarray
        Vector of rmin parameters, in angstoms,  corresponding to atoms pairs of r.
    eps : numpy.ndarray
        Vector of epsilon parameters corresponding to atom pairs of r
    forgive : numpy.ndarray
        The `forgive` factor is a softening term to mitigate rigid body artifacts. It is set to a value between 0 1nd 1
        and modifies the rmin parameter of all atom pairs in `r` to be the fraction, `forgive` of the thair original
        value. This allows atoms to be closer than otherwise allowed to prevent explosion of the jennard-jones repulsion
        in situations that would otherwise be resolved with minor atomic displacements.
    cap : flaot
        Maximum allowed energy factor. Sets a cap on the maximum energy contribution of one atom pair interaction as
        another mechanism for softening. `cap` is sets the actual max value as a multiple of the `eps` parameter. i.e.
        the maximum allowed energy from a single atom pair interaction is the `eps` parameter multiplied by `cap`.

    Returns
    -------
    lj_energy: numpy.ndarray
        Vector of atom pair energies calculated using the modified lj potential function.
    """
    r = np.atleast_2d(r)
    lj_energy = np.empty_like(r)
    rmin_lower = forgive * rmin

    # Piecewise function for flat lj potential near rmin
    for i, rotamer in enumerate(r):
        for j in range(len(rotamer)):
            lj = rmin_lower[j] / rotamer[j]
            lj = lj * lj * lj
            lj = lj * lj
            lj_energy[i, j] = np.minimum(eps[j] * lj**2, cap * eps[j])

    return lj_energy

@clash_only
@njit(cache=True)
def get_lj_attr(r, rmin, eps, forgive=0.9, floor=-2):
    """Calculate only the attractive terms of the lenard-jones potential from a set of atom pairs with distance `r`,
      rmin values of `rmin` and epsilon values of `eps`.

      Parameters
      ----------
      r : numpy.ndarray
          Vector of inter-atomic distances between non-bonded atoms of a system.
      rmin : numpy.ndarray
          Vector of rmin parameters, in angstoms,  corresponding to atoms pairs of r.
      eps : numpy.ndarray
          Vector of epsilon parameters corresponding to atom pairs of r
      forgive : numpy.ndarray
          The `forgive` factor is a softening term to mitigate rigid body artifacts. It is set to a value between 0 1nd 1
          and modifies the rmin parameter of all atom pairs in `r` to be the fraction, `forgive` of the thair original
          value. This allows atoms to be closer than otherwise allowed to prevent explosion of the jennard-jones repulsion
          in situations that would otherwise be resolved with minor atomic displacements.
      cap : flaot
          Maximum allowed energy factor. Sets a cap on the maximum energy contribution of one atom pair interaction as
          another mechanism for softening. `cap` is sets the actual max value as a multiple of the `eps` parameter. i.e.
          the maximum allowed energy from a single atom pair interaction is the `eps` parameter multiplied by `cap`.

      Returns
      -------
      lj_energy: numpy.ndarray
          Vector of atom pair energies calculated using the modified lj potential function.
      """
    r = np.atleast_2d([r])
    lj_energy = np.empty_like(r)
    rmin_lower = forgive * rmin
    for j, rotamer in enumerate(r):
        # Piecewise function for flat lj potential near rmin
        for i in range(len(rotamer)):
            lj = rmin_lower[i] / rotamer[i]
            lj = lj * lj * lj
            lj = lj * lj
            lj_energy[j, i] = np.maximum(-2 * eps[i] * lj, eps[i] * floor)

    return lj_energy


def prep_external_clash(ensemble):
    """ Helper function to prepare the lj parameters of a rotamer ensemble, presumably with an associated protein.

    Parameters
    ----------
    ensemble : RotamerEnsemble
        The RotamerEnsemble to prepare the clash parameters for. The RotamerEnsemble should have an associated protein.

    Returns
    -------
        dist : numpy.ndarray
            Array of pairwise distances between atoms in the RotamerEnsemble and atoms in the associated protein.
        rmin_ij :  numpy.ndarray
            ``rmin`` parameters of the lj potential associated with the atom ``i`` and atom ``j`` pair.
        eps_ij : numpy.ndarray
            ``eps`` parameters of the lj potential associated with the atom ``i`` and atom ``j`` pair.
        shape : Tuple[int]
            Shape the array should be so that the energy is evaluated for each rotamer of the ensemble separately.
    """

    # Calculate rmin and epsilon for all atoms in protein that may clash
    if hasattr(ensemble, 'ermin_ij'):
        rmin_ij = ensemble.ermin_ij
        eps_ij = ensemble.eeps_ij
    else:
        rmin_ij, eps_ij = get_lj_params(ensemble)
        ensemble.ermin_ij = rmin_ij
        ensemble.eeps_ij = eps_ij

    if len(ensemble) == 1:
        ec = ensemble.coords[0, ensemble.side_chain_idx]
    else:
        ec = ensemble.coords[:, ensemble.side_chain_idx].reshape(-1, 3)
    pc = ensemble.protein_tree.data[ensemble.protein_clash_idx]
    # Calculate distances
    dist = cdist(ec, pc).reshape(len(ensemble), -1)
    return dist, rmin_ij, eps_ij


def prep_internal_clash(ensemble):
    """ Helper function to prepare the lj parameters of a rotamer ensemble to evaluate internal clashes.

        Parameters
        ----------
        ensemble : RotamerEnsemble
            The RotamerEnsemble to prepare the clash parameters for.

        Returns
        -------
            dist : numpy.ndarray
                Array of pairwise distances between atoms in the RotamerEnsemble and atoms in the associated protein.
            rmin_ij :  numpy.ndarray
                ``rmin`` parameters of the lj potential associated with the atom ``i`` and atom ``j`` pair.
            eps_ij : numpy.ndarray
                ``eps`` parameters of the lj potential associated with the atom ``i`` and atom ``j`` pair.
            shape : Tuple[int]
                Shape the array should be so that the energy is evaluated for each rotamer of the ensemble separately.
        """

    if hasattr(ensemble, 'ermin_ij'):
        rmin_ij, eps_ij = ensemble.ermin_ij, ensemble.eeps_ij
    else:
        a, b = [list(x) for x in zip(*ensemble.non_bonded)]
        a_eps = chilife.get_lj_eps(ensemble.atom_types[a])
        a_radii = chilife.get_lj_rmin(ensemble.atom_types[a])
        b_eps = chilife.get_lj_eps(ensemble.atom_types[b])
        b_radii = chilife.get_lj_rmin(ensemble.atom_types[b])

        join_rmin = chilife.get_lj_rmin("join_protocol")[()]
        join_eps = chilife.get_lj_eps("join_protocol")[()]

        rmin_ij = join_rmin(a_radii * ensemble.forgive, b_radii * ensemble.forgive, flat=True)
        eps_ij = join_eps(a_eps, b_eps, flat=True)

        ensemble.irmin_ij = rmin_ij
        ensemble.ieps_ij = eps_ij
        ensemble.aidx = a
        ensemble.bidx = b

    dist = np.linalg.norm(ensemble.coords[:, ensemble.aidx] - ensemble.coords[:, ensemble.bidx], axis=2)
    shape = (len(ensemble.coords), len(a_radii))

    return dist, rmin_ij, eps_ij, shape


def reweight_rotamers(energies, temp, weights):
    """Adjust rotamer population weights based on external energies (from clash evaluations).

    Parameters
    ----------
    energies : numpy.ndarray
        Array of external energies, in kcal/mol
    temp : scalar
        Temperature, in kelvin
    weights : numpy.ndarray
        Current weights of rotamers
    return_partition : bool
        If True, return the value of the partition function an additional argument; see below

    Returns
    -------
    new_weights : numpy.ndarray
        Adjusted weights
    partition : float (optional)
        The partition function relative to the free label. A small partition function suggests the interactions with
        neighboring atoms are unfavorable while a large partition function suggests the opposite.
    """

    probabilities = np.exp(-energies / (chilife.GAS_CONST * temp))
    p = probabilities * weights
    p_sum = np.sum(p)
    new_weights = p/p_sum
    partition = p_sum / weights.sum()

    return new_weights, partition


def get_lj_params(ensemble):
    """ calculate the lennard jones parameters between atoms of a rotamer ensemble and associated protein.

    Parameters
    ----------
    ensemble : RotamerEnsemble
        The RotamerEnsemble to get the lj params for. Should have an associated protein object.

    Returns
    -------
    rmin_ij : nummpy.ndarray
        Vector of ``rmin`` lennard-jones parameters corresponding to atoms i and j of the RotamerEnsemble and the
        associated protein respectively.
    eps_ij : numpy.ndarray
        Vector of ``eps`` lennard-jones parameters corresponding to atoms i and j of the RotamerEnsemble and the
        associated protein respectively.
    """
    environment_atypes = ensemble.protein.atoms.types[ensemble.protein_clash_idx]
    protein_lj_radii = chilife.get_lj_rmin(environment_atypes)
    protein_lj_eps = chilife.get_lj_eps(environment_atypes)
    join_rmin = chilife.get_lj_rmin("join_protocol")[()]
    join_eps = chilife.get_lj_eps("join_protocol")[()]

    rmin_ij = join_rmin(ensemble.rmin2 * ensemble.forgive, protein_lj_radii * ensemble.forgive).reshape(-1)
    eps_ij = join_eps(ensemble.eps, protein_lj_eps).reshape((-1))

    return rmin_ij, eps_ij


def join_geom(a, b, flat=False):
    """ Function to join Lennard-Jones parameters (``rmin`` or ``eps``) using their geometric mean. Parameters can be
    joined in two different ways (see keyword argument ``flat``)

    Parameters
    ----------
    a : numpy.ndarray
        Single atom parameters of atoms of group a
    b : numpy.ndarray
        Single atom parameters of atoms of group b
    flat : bool
        Only join parameters i of group a and j of group b if i=j. If false all combinations of i,j are computed.

    Returns
    -------
    param : numpy.ndarray
        The joined parameters of group a and b
    """

    if flat:
        return np.sqrt(a * b)
    else:
        return np.sqrt(np.outer(a, b))


def join_arith(a, b, flat=False):
    """ Function to join Lennard-Jones parameters (``rmin`` or ``eps``) using their arithmatic mean. Parameters can be
    joined in two different ways (see keyword argument ``flat``)

    Parameters
    ----------
    a : numpy.ndarray
        Single atom parameters of atoms of group a
    b : numpy.ndarray
        Single atom parameters of atoms of group b
    flat : bool
        Only join parameters i of group a and j of group b if i=j. If false all combinations of i,j are computed.

    Returns
    -------
    param : numpy.ndarray
        The joined parameters of group a and b
    """
    if flat:
        return a + b
    else:
        return np.add.outer(a, b)


rmin_charmm = {
     "C": 2.0000,
     "H": 1.27000,
     "N": 1.8500,
     "O": 1.7000,
     "S": 2.0000,
    "Se": 2.0000,  # Default Selenium to Sulfur
    "Br": 1.9800,
    "Cu": 0.7251,  # ion 2+
     'B': 2.5500,
     'F': 1.6300,
    'Na': 1.3638,  # ion
    'Mg': 1.1850,
     'P': 2.1500,
    'Cl': 1.9100,
     'K': 1.7638,  # ion
    'Ca': 1.3670,  # ion
    'Fe': 0.6500,  # HEM
    'Ni': 1.2760,
    'Zn': 1.0900,  # ion
    'Gd': 1.5050,  # ion 3+
    "join_protocol": join_arith,
}  # Bromobenzene from Gutiérrez et al. 2016

eps_charmm = {
    "C": -0.110,
    "H": -0.022,
    "N": -0.200,
    "O": -0.120,
    "S": -0.450,
    "Se": -0.450,
    "Br": -0.320,
    "Cu": -0.1505,
     'B': -0.0380,
     'F': -0.105,
    'Na': -0.0469,
    'Mg': -0.0150,
     'P': -0.5850,
    'Cl': -0.3430,
     'K': -0.0870,
    'Ca': -0.120,
    'Fe': 0.000,
    'Ni': -5.65,
    'Zn': -0.250,
    'Gd': -0.1723,
    "join_protocol": join_geom,
}  # Bromobenzene from Gutiérrez et al. 2016

rmin_uff = {
    "C": 3.851,
    "H": 2.886,
    "N": 3.660,
    "O": 3.500,
    "S": 4.035,
    "Se": 4.035,  # Default Selenium to Sulfur
    "Br": 4.189,
    "Na": 2.983,
    "Mg": 3.021,
    "P" : 4.147,
    "Cl": 3.947,
    "K" : 3.812,
    "Ca": 3.399,
    "Mn": 2.961,
    "Fe": 2.912,
    "Co": 2.872,
    "Ni": 2.834,
    "Cu": 3.495,
    "Gd": 3.368,
    "join_protocol": join_geom,
}

eps_uff = {
    "C": -0.105,
    "H": -0.044,
    "N": -0.069,
    "O": -0.060,
    "S": -0.274,
    "Se": -0.274,  # Default Selenium to Sulfur
    "Br": -0.251,
    "Na": -0.030,
    "Mg": -0.111,
    "P" : -0.305,
    "Cl": -0.227,
    "K" : -0.035,
    "Ca": -0.238,
    "Mn": -0.013,
    "Fe": -0.013,
    "Co": -0.014,
    "Ni": -0.015,
    "Cu": -0.005,
    "Gd": -0.009,
    "join_protocol": join_geom,
}

for dictionary in rmin_charmm, eps_charmm, rmin_uff, eps_uff:
    tdict = {**{key.upper(): val for key, val in dictionary.items()},
             **{key.lower(): val for key, val in dictionary.items()}}
    dictionary.update(tdict)

lj_params = {"uff": [rmin_uff, eps_uff], "charmm": [rmin_charmm, eps_charmm]}


def set_lj_params(forcefield):
    """
        Global setting for using different atom type parameterization for clash evaluations. Running this function
        once will set the parameters for the whole session.

    Parameters
    ----------
    forcefield : str
        Name of the forcefield to be used.

    """
    chilife.using_lj_param = forcefield
    rmin_func, eps_func = lj_params[forcefield]
    chilife.get_lj_rmin = np.vectorize(rmin_func.__getitem__)
    chilife.get_lj_eps = np.vectorize(eps_func.__getitem__)


set_lj_params("charmm")

KCAL2J = 4.184e3  # conversion factor form kcal to J (exact)
BOLTZ_CONST = 1.380649e-23  # Boltzmann constant, J K^-1 (exact)
AVOGADRO_CONST = 6.02214076e23  # Avogadro constant, mol^-1 (exact)
GAS_CONST = BOLTZ_CONST*AVOGADRO_CONST/KCAL2J  # molar gas constant, kcal mol^1 K^-1



from .chilife import *
from .io import *
from .alignment_methods import parse_backbone
from .RotamerEnsemble import *
from .SpinLabel import *
from .dSpinLabel import dSpinLabel
from .IntrinsicLabel import IntrinsicLabel
from .SpinLabelTraj import *
from .Protein import *
from .ProteinIC import *
from .protein_utils import *
from .scoring import *



import numpy as np
from scipy.stats import norm
from scipy.special import kl_div
from scipy.spatial.distance import cdist
import pytest
import MDAnalysis as mda
import chilife.numba_utils as nu
import chilife as xl


def test_compute_bin():
    bin_edges = np.linspace(0, 10, 5)
    var = 5
    assert nu.compute_bin(var, bin_edges) == 2

    var = 50
    assert nu.compute_bin(var, bin_edges) == 20


def test_dirichlet():
    alphas = np.array([0.1, 0.2, 0.3, 0.4], dtype=float)

    np.random.seed(100000)
    Ans = np.random.dirichlet(alphas, 1000000).mean(axis=0)
    Test = np.array([nu.dirichlet(alphas) for i in range(1000000)]).mean(axis=0)
    np.testing.assert_almost_equal(Test, Ans, decimal=3)


def test_get_delta_r():
    x = np.arange(0, 100, 10, dtype=float)
    assert nu.get_delta_r(x) == 9.0


def test_histogram():
    r = np.linspace(0, 100, 2**8)
    data = np.random.normal(45, 5, 100)
    w = np.random.uniform(0, 1, 100)
    w /= w.sum()

    np_hist, _ = np.histogram(data, r, weights=w)
    nu_hist = nu.histogram(data, w, r)

    np.testing.assert_almost_equal(np_hist, nu_hist[:-1])


def test_jaccard():
    y1 = np.zeros(10)
    y1[3:7] = 1

    y2 = np.zeros(10)
    y2[4:8] = 1

    assert nu.jaccard(y1, y2) == 3 / 5


def test_kl_divergence():
    r = np.linspace(0, 100, 2**8)
    p1 = norm(45, 5).pdf(r)
    p2 = norm(55, 6).pdf(r)

    assert nu.kl_divergence(p1, p2) == pytest.approx(kl_div(p1, p2).sum())


def test_normdist():
    mu, sigma = 50, 10
    delta_r = 1e-2
    x = np.arange(mu - 3.5 * sigma, mu + 3.5 * sigma + delta_r, delta_r)

    y1 = norm(mu, sigma).pdf(x)
    y2 = nu.normdist(delta_r, 50, 10)[1]

    np.testing.assert_almost_equal(y1, y2)


def test_pairwise_dist():
    from time import time

    A = np.random.rand(1000, 3)
    B = np.random.rand(1000, 3)
    print(A.shape)
    nu.pairwise_dist(A, B)

    t1 = time()
    D_cdist = cdist(A, B)
    t2 = time()
    D_numba = nu.pairwise_dist(A, B)
    t3 = time()
    print("cdist time: ", t2 - t1)
    print("numba time: ", t3 - t2)

    np.testing.assert_almost_equal(D_cdist, D_numba)


def test_fib_points():
    x = nu.fibonacci_points(10)
    ans = np.array([[ 0.43588989,  0.        ,  0.9],
                    [-0.52658671,  0.48239656,  0.7],
                    [ 0.0757129 , -0.86270943,  0.5],
                    [ 0.58041368,  0.75704687,  0.3],
                    [-0.97977755, -0.17330885,  0.1],
                    [ 0.83952592, -0.53403767, -0.1],
                    [-0.24764672,  0.92123347, -0.3],
                    [-0.39915719, -0.76855288, -0.5],
                    [ 0.67080958,  0.24497858, -0.7],
                    [-0.40291289,  0.16631658, -0.9]])

    np.testing.assert_allclose(x, ans)


def test_get_sasa1():
    ubq = mda.Universe('test_data/1ubq.pdb').select_atoms('protein')
    SL = xl.SpinLabel('R1M', 28, ubq)
    atom_coords = SL.coords[0]
    atom_radii = xl.get_lj_rmin(SL.atom_types)

    environment_coords = SL.protein.atoms[SL.protein_clash_idx].positions
    environment_radii = xl.get_lj_rmin(SL.protein.atoms[SL.protein_clash_idx].types)

    area = nu.get_sasa(atom_coords, atom_radii, environment_coords, environment_radii)

    assert area[0] == 272.54997512834973


def test_get_sasa2():
    ubq = mda.Universe('test_data/1ubq.pdb').select_atoms('protein')

    atom_coords = ubq.atoms.positions
    atom_radii = xl.get_lj_rmin(ubq.atoms.types)

    area = nu.get_sasa(atom_coords, atom_radii, by_atom=True)



    assert area.sum() == 4832.089121458064

    ans = np.array([2.83037863, 27.47825921, 0., 0., 0.,
                    0., 0., 0.14186254, 0., 30.89830006])

    np.testing.assert_allclose(area[0, 300:310], ans)





import logging
from copy import deepcopy
import numpy as np
import chilife
from .dRotamerEnsemble import dRotamerEnsemble


class dSpinLabel(dRotamerEnsemble):
    def __init__(self, label, sites, protein=None, chain=None, rotlib=None, **kwargs):
        super().__init__(label, sites, protein=protein, chain=chain, rotlib=rotlib, **kwargs)
        self.label = self.res

    def create_ensembles(self):
        self.SL1 = chilife.SpinLabel(self.res,
                                     self.site1,
                                     self.protein,
                                     self.chain,
                                     self.libA,
                                     **self.kwargs)

        self.SL2 = chilife.SpinLabel(self.res,
                                     self.site2,
                                     self.protein,
                                     self.chain,
                                     self.libB,
                                     **self.kwargs)

        self.RE1, self.RE2 = self.SL1, self.SL2

    def copy(self):
        new_copy = chilife.dSpinLabel(self.res, (self.site1, self.site2), chain=self.chain,
                                      protein=self.protein,
                                      rotlib={'csts': self.csts, 'libA': self.libA, 'libB': self.libB},
                                      minimize=False,
                                      eval_clash=False)
        for item in self.__dict__:
            if isinstance(self.__dict__[item], np.ndarray):
                new_copy.__dict__[item] = self.__dict__[item].copy()

            elif item == 'SL1':
                new_copy.__dict__[item] == self.__dict__[item].copy(rotlib=self.libA)
                new_copy.__dict__['RE1'] == new_copy.__dict__[item]

            elif item == 'SL2':
                new_copy.__dict__[item] == self.__dict__[item].copy(rotlib=self.libB)
                new_copy.__dict__['RE2'] = new_copy.__dict__[item]

            elif item in ('protein', 'RE1', 'RE2'):
                pass

            else:
                new_copy.__dict__[item] = deepcopy(self.__dict__[item])

        return new_copy

    @property
    def spin_atoms(self):
        return np.unique(np.concatenate((self.SL1.spin_atoms, self.SL2.spin_atoms)))

    @property
    def spin_idx(self):
        return np.argwhere(np.isin(self.atom_names, self.spin_atoms)).flatten()

    @property
    def spin_coords(self):
        return self.coords[:, self.spin_idx]

    @property
    def spin_centers(self):
        return np.average(self.spin_coords, axis=1, weights=self.spin_weights)

    @property
    def spin_weights(self):
        return self.SL1.spin_weights

    @property
    def spin_centroid(self):
        return np.average(self.spin_coords, weights=self.weights, axis=0)



import numpy as np
import pytest
import MDAnalysis as mda
import chilife

import matplotlib.pyplot as plt

r = np.linspace(1e-3, 3, 256)
eps = np.ones(len(r))
rmin = np.ones(len(r))
protein = mda.Universe("test_data/1ubq.pdb", in_memory=True)
lj_funcs = [chilife.get_lj_energy, chilife.get_lj_scwrl, chilife.get_lj_rep]
lj_ans = [
    [-1.2940752, -1.34917751, -0.92307548, 0.09701376, -0.81320862, -0.82519124, -0.8168207, -0.47411124, -0.64444488, -0.44549325],
    [-0.05528246, -0.14662195, 0., -0.20340994, 0., 0., 0., 0., 0., 0.],
    [0.66691997, 0.05328764, 0.01843083, 0.01921863, 0.00483313, 0.68429553, 0.00290177, 0.00411807, 1.25053418, 0.03108684],
]


@pytest.mark.parametrize(('func', 'ans'), zip(lj_funcs, lj_ans))
def test_lj(func, ans):
    RL = chilife.RotamerEnsemble('TRP', 28, protein, energy_func=func, eval_clash=True)
    np.testing.assert_almost_equal(RL.atom_energies[5], ans)


@pytest.mark.parametrize('func',  lj_funcs)
def test_efunc(func):
    RL = chilife.RotamerEnsemble('TRP', 28, protein)
    test = func(RL)
    ans = np.load(f'test_data/{func.__name__}.npy')
    np.testing.assert_almost_equal(test, ans)


@pytest.mark.parametrize('func',  lj_funcs)
def test_efunc_dlabel(func):
    dSL = chilife.dSpinLabel('DHC', (28, 32), protein, eval_clash=False)
    test = func(dSL)
    ans = np.load(f'test_data/d{func.__name__}.npy')
    np.testing.assert_almost_equal(test, ans, decimal=4)


def test_prep_internal_clash():
    SL = chilife.RotamerEnsemble('TRP', 28, protein)
    r, rmin, eps, shape = chilife.scoring.prep_internal_clash(SL)
    with np.load('test_data/internal_clash_prep.npz') as f:
        r_ans, rmin_ans, eps_ans = f['r'], f['rmin'], f['eps']

    assert shape == (36, 56)
    np.testing.assert_almost_equal(r, r_ans)
    np.testing.assert_almost_equal(rmin, rmin_ans)
    np.testing.assert_almost_equal(eps, eps_ans)


def test_molar_gas_constant():
    np.testing.assert_almost_equal(chilife.scoring.GAS_CONST, 1.987204258640832e-3, decimal=10)


def test_get_lj_case_sensitivity():
    x = chilife.get_lj_rmin(['CA', 'Ca', 'ca'])
    assert np.all(x == 1.367)


import hashlib, os, pickle
import pytest
import numpy as np
import MDAnalysis as mda
import chilife as xl

pdbids = ["1ubq", "1a2w", '1az5']
ubq = mda.Universe("test_data/1ubq.pdb", in_memory=True)
ICs = xl.get_internal_coords(ubq)

gd_kwargs = [
    {"resi": 28, "atom_list": ["C", "N", "CA", "C"]},
    {"resi": 28, "atom_list": [["C", "N", "CA", "C"], ["N", "CA", "C", "N"]]},
]
gd_ans = [-1.1540443794802524, np.array([-1.15404438, -0.66532042])]


def test_save_pdb():
    protein = mda.Universe("test_data/alphabetical_peptide.pdb").select_atoms("protein")

    uni_ics = xl.get_internal_coords(protein)
    uni_ics.save_pdb("test_data/postwrite_alphabet_peptide.pdb")

    with open("test_data/postwrite_alphabet_peptide.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("test_data/alphabetical_peptide.pdb", "r") as f:
        truth = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove("test_data/postwrite_alphabet_peptide.pdb")
    assert test == truth


def test_to_site():
    backbone = ubq.select_atoms("resnum 28 and name N CA C").positions
    r1c = mda.Universe("../chilife/data/rotamer_libraries/residue_pdbs/R1C.pdb")
    R1ic = xl.get_internal_coords(r1c)
    R1ic.to_site(*backbone)

    np.testing.assert_almost_equal(R1ic.coords[1], backbone[1])


def test_has_clashes():
    assert not ICs.has_clashes()
    ICs.set_dihedral(np.pi / 2, 35, ["N", "CA", "C", "N"])
    assert ICs.has_clashes()


@pytest.mark.parametrize("pdbid", pdbids)
def test_get_internal_coordinates(pdbid):
    protein = mda.Universe(f"test_data/{pdbid}.pdb", in_memory=True).select_atoms("protein and not altloc B")
    ICs = xl.get_internal_coords(protein)

    np.testing.assert_almost_equal(ICs.coords, protein.atoms.positions, decimal=4)

def test_set_dihedral():
    ICs = xl.get_internal_coords(ubq)
    ICs.set_dihedral(
        [-np.pi / 2, np.pi / 2], 72, [["C", "N", "CA", "C"], ["N", "CA", "C", "N"]]
    )

    ans = mda.Universe("test_data/test_icset.pdb")

    np.testing.assert_almost_equal(ans.atoms.positions, ICs.coords, decimal=3)

def test_set_dihedral2():
    lys = mda.Universe('../chilife/data/rotamer_libraries/residue_pdbs/lys.pdb')

    ICs = xl.get_internal_coords(lys)
    ICs.set_dihedral(np.pi / 2, 1, ["N", "CA", "CB", "CG"])

    print('DSF')


@pytest.mark.parametrize(["inp", "ans"], zip(gd_kwargs, gd_ans))
def test_get_dihedral(inp, ans):
    dihedral = ICs.get_dihedral(**inp)
    np.testing.assert_almost_equal(dihedral, ans)



def test_polypro():
    polypro = mda.Universe("test_data/PPII_Capped.pdb")
    polyproIC = xl.get_internal_coords(polypro)


def test_PRO():
    pro = mda.Universe("../chilife/data/rotamer_libraries/residue_pdbs/pro.pdb")
    pro_ic = xl.get_internal_coords(pro)
    assert ("CD", "CG", "CB", "CA") in pro_ic.ICs[1][1]

def test_PRO2():
    ubq_IC = xl.get_internal_coords(ubq)

    assert ("C", "CA", "N", "C") in ubq_IC.ICs[1][37]
    assert ("C", "CA", "N", "C") in ubq_IC.ICs[1][38]


def test_set_coords():
    R1A = mda.Universe("test_data/R1A.pdb")
    R1A_IC = xl.get_internal_coords(R1A)
    R1A_IC_c = R1A_IC.copy()
    R1A_IC_c.set_dihedral([np.pi/2, -np.pi/2, np.pi/2], 1, [['N', 'CA', 'CB', 'SG' ],
                                                            ['CA', 'CB', 'SG', 'SD'],
                                                            ['CB', 'SG', 'SD', 'CE']])

    coords = R1A_IC_c.coords

    R1A_IC.coords = coords
    np.testing.assert_allclose(R1A_IC.zmats[1], R1A_IC_c.zmats[1], rtol=1e-6)
    np.testing.assert_almost_equal(R1A_IC.coords, R1A_IC_c.coords, decimal=6)


def test_nonbonded():
    with open('test_data/ic_nb.pkl', 'rb') as f:
        nb_ans = pickle.load(f)

    np.testing.assert_equal(ICs.nonbonded_pairs, nb_ans)


def test_get_zmat_idxs():
    R1A = mda.Universe("test_data/R1A.pdb")
    R1A_IC = xl.get_internal_coords(R1A)
    idxs, stem, idx = R1A_IC.get_zmat_idxs(1, ['CB', 'SG', 'SD', 'CE'], 1)
    assert stem == ('SD', 'SG', 'CB')
    assert idx == -1

    idxs, stem, idx = R1A_IC.get_zmat_idxs(1, ['CE', 'SD', 'SG', 'CB'], 1)
    assert stem == ('SD', 'SG', 'CB')
    assert idx == 0

def test_phi_idxs():

    # Still need to test chain

    idxs = ICs.phi_idxs(range(4, 11,))
    vals = ICs.zmats[1][idxs, -1]
    ans = [ICs.get_dihedral(i, ('C', 'N', 'CA', 'C')) for i in range(4, 11)]

    np.testing.assert_almost_equal(vals, ans)

    idx = ICs.phi_idxs(1)
    assert len(idx) == 0


def test_psi_idxs():

    # Still need to test chain

    idxs = ICs.psi_idxs(range(4, 11,))
    vals = ICs.zmats[1][idxs, -1]
    ans = [ICs.get_dihedral(i, ('N', 'CA', 'C', 'N')) for i in range(4, 11)]

    np.testing.assert_almost_equal(vals, ans)

    idx = ICs.psi_idxs(76)
    assert len(idx) == 0

def test_phi_psi_idxs_multichain():
    prot = xl.fetch('1a2w').select_atoms('protein')
    ICs = xl.get_internal_coords(prot)

    with pytest.raises(ValueError):
        ICs.psi_idxs(10)

    with pytest.raises(ValueError):
        ICs.psi_idxs(10)

    A = ICs.psi_idxs(10, chain=2)
    B = ICs.phi_idxs(10, chain=2)

    assert A[0] == 80
    assert B[0] == 71

def test_chi_idxs():
    idxs = ICs.chi_idxs(range(4, 11, ))
    idxs = np.concatenate(idxs)

    vals = ICs.zmats[1][idxs, -1]
    ans = np.array([-1.0589305, -3.1378433, -3.0641198,  1.338909 , -1.1897472,
                     1.2340594,  1.7094074,  3.0752275, -2.8944212, -3.0027205,
                    -3.040881 ])

    np.testing.assert_almost_equal(vals, ans)

def test_ic_pref_dihe():

    mol = mda.Universe('test_data/test_ic_pref_dihe.pdb', in_memory=True)
    dih = [['N', 'CA', 'CB', 'CB2'],
           ['CA', 'CB', 'CB2', 'NG'],
           ['ND', 'CE3', 'CZ3', 'C31']]

    IC = xl.get_internal_coords(mol, preferred_dihedrals=dih)
    IC.set_dihedral(np.pi / 2, 1, ['ND', 'CE3', 'CZ3', 'C31'])

    sister_dihe_atom_coords = IC.coords[IC.atom_names == 'C36'].flat
    np.testing.assert_almost_equal( sister_dihe_atom_coords, [2.14953486, -3.40647599, -0.93880627])



import numpy as np
from .protein_utils import FreeAtom


class IntrinsicLabel:

    def __init__(self, res, atom_selection, spin_atoms=None, name='IntrinsicLabel'):
        """
        A helper object to assign spin density to part of a protein that is already present, e.g. metal ligands.
        IntrinsicLabels can be used with the :func:`~chilife.chilife.distance_distribution` function

        Parameters
        ----------
        res : str
            3-character identifier of desired residue, e.g. NC2 (Native Cu(II)).
        atom_selection : MDAnalysis.AtomGroup, chiLife.System
            Group of atoms constituting the intrinsic label. These selections can consist of multiple residues, which
            can be useful in the case of ions with multiple coordinating residues

        spin_atoms : str, list, tuple, array, dict, MDAnalysis.AtomGroup
            Atoms of the intrinsic label that host the unpaired electron density. Can be a single atom name, a
            list/tuple/array of atom names or a dictionary mapping atom names to their relative populations.
            ``spin_atoms `` can also be an ``MDAnalysis.AtomGroup`` object derived from the same MDAnalysis.Universe
            as the ``atom_selection`` keyword argument. Use of an ``AtomGroup`` is particularly useful when there is
            spin density is distributed on several atoms with the same name on different residues within the
            ``IntrinsicLabel``.
        """
        self._selection = atom_selection
        self._coords = atom_selection.positions.copy()[None, :, :]
        self.coords = self._coords
        self.atom_names = atom_selection.names.copy()
        self.atom_types = atom_selection.types
        self.name = name
        self.label = res
        self.res = res


        self.chain = "".join(set(atom_selection.segids))
        self.weights = np.array([1])
        if isinstance(spin_atoms, (list, tuple, np.ndarray)):
            self.spin_atoms = np.asarray(spin_atoms)
            self.spin_weights = np.ones(len(spin_atoms)) / len(spin_atoms)
        elif isinstance(spin_atoms, dict):
            self.spin_atoms = np.asarray([key for key in spin_atoms])
            self.spin_weights = np.asarray([val for val in spin_atoms.values()])
        elif isinstance(spin_atoms, str):
            self.spin_atoms = np.asarray([spin_atoms])
            self.spin_weights = np.asarray([1])
        else:
            raise RuntimeError("spin_atoms must contain a string, a list/tuple/array of strings, a dict")

        self.site = atom_selection.select_atoms(f'name {self.spin_atoms[np.argmax(self.spin_weights)]}').resnums[0]
        sa_mask = np.isin(self.atom_names, self.spin_atoms,)

        self.spin_idx = np.argwhere(sa_mask)
        self.spin_idx.shape = -1

        self.atoms = [FreeAtom(atom.name, atom.type, idx, atom.resname, atom.resnum, atom.position)
                      for idx, atom in enumerate(self._selection.atoms)]

    @property
    def spin_coords(self):
        """get the spin coordinates of the rotamer ensemble"""
        return np.squeeze(self._coords[:, self.spin_idx, :])

    @property
    def spin_centers(self):
        """get the spin center of the rotamers in the ensemble"""
        if np.any(self.spin_idx):
            spin_centers = np.average(self._coords[:, self.spin_idx, :], weights=self.spin_weights, axis=1)
        else:
            spin_centers = np.array([])
        return np.atleast_2d(np.squeeze(spin_centers))

    @property
    def spin_centroid(self):
        """Average location of all the label's `spin_coords` weighted based off of the rotamer weights"""
        return np.average(self.spin_centers, weights=self.weights, axis=0)



from copy import deepcopy
from functools import partial
import numpy as np
from scipy.spatial.distance import cdist
from scipy.spatial import cKDTree
import MDAnalysis as mda
from .RotamerEnsemble import RotamerEnsemble
import chilife


class SpinLabel(RotamerEnsemble):
    """
    A RotamerEnsemble made from a side chain with one or more unpaired electrons.

    :param label: str
        Name of desired spin label, e.g. R1A.

    :param site: int
        Protein residue number to attach spin label to.

    :param chain: str
        Protein chain identifier to attach spin label to.

    :param protein: MDAnalysis.Universe, MDAnalysis.AtomGroup
        Object containing all protein information (coords, atom types, etc.)

    :param protein_tree: cKDtree
        k-dimensional tree object associated with the protein coordinates.
    """

    def __init__(self, label, site=None, protein=None, chain=None, rotlib=None, **kwargs):

        # Override RotamerEnsemble default of not evaluating clashes
        if not kwargs.get('minimize', False):
            kwargs.setdefault("eval_clash", True)

        super().__init__(label, site, protein=protein, chain=chain, rotlib=rotlib, **kwargs)

        self.label = label

        # Parse spin delocalization information
        sa_mask = np.isin(self.spin_atoms, self.atom_names)
        self.spin_atoms = self.spin_atoms[sa_mask]
        self.spin_weights = self.spin_weights[sa_mask]
        self.spin_idx = np.argwhere(np.isin(self.atom_names, self.spin_atoms))
        self.spin_idx.shape = -1

    @property
    def spin_coords(self):
        """get the spin coordinates of the rotamer ensemble"""
        return np.squeeze(self._coords[:, self.spin_idx, :])

    @property
    def spin_centers(self):
        """get the spin center of the rotamers in the ensemble"""
        if np.any(self.spin_idx):
            spin_centers = np.average(self._coords[:, self.spin_idx, :], weights=self.spin_weights, axis=1)
        else:
            spin_centers = np.array([])
        return np.atleast_2d(np.squeeze(spin_centers))

    @property
    def spin_centroid(self):
        """Average location of all the label's `spin_coords` weighted based off of the rotamer weights"""
        return np.average(self.spin_centers, weights=self.weights, axis=0)

    def protein_setup(self):

        if isinstance(self.protein, (mda.AtomGroup, mda.Universe)):
            if not hasattr(self.protein.universe._topology, "altLocs"):
                self.protein.universe.add_TopologyAttr('altLocs', np.full(len(self.protein.universe.atoms), ""))

        self.protein = self.protein.select_atoms("not (byres name OH2 or resname HOH)")
        self.to_site()
        self.backbone_to_site()
        self.clash_ignore_idx = self.protein.select_atoms(
            f"resid {self.site} and segid {self.chain}"
        ).ix

        self.resindex = self.protein.select_atoms(self.selstr).resindices[0]
        self.segindex = self.protein.select_atoms(self.selstr).segindices[0]

        if self.protein_tree is None:
            self.protein_tree = cKDTree(self.protein.atoms.positions)

        protein_clash_idx = self.protein_tree.query_ball_point(
            self.clash_ori, self.clash_radius
        )
        self.protein_clash_idx = [
            idx for idx in protein_clash_idx if idx not in self.clash_ignore_idx
        ]

        # Evaluate external clash energies and reweight rotamers
        if self._minimize and self.eval_clash:
            raise RuntimeError('Both `minimize` and `eval_clash` options have been selected, but they are incompatible.'
                               'Please select only on. Also note that minimize performs its own clash evaluations so '
                               'eval_clash is not necessary.')
        elif self.eval_clash:
            self.evaluate()

        elif self._minimize:
            self.minimize()

    @classmethod
    def from_mmm(cls, label, site, protein=None, chain=None, **kwargs):
        """Create a SpinLabel object using the default MMM protocol with any modifications passed via kwargs"""

        MMM_maxdist = {
            "R1M": 9.550856367392733,
            "R7M": 9.757254987175209,
            "V1M": 8.237071322458029,
            "M1M": 8.985723827323680,
            "I1M": 12.952083029729994,
        }

        # Store the force field parameter set being used before creating the spin label
        curr_lj = chilife.using_lj_param
        user_lj = kwargs.pop("lj_params", "uff")
        # Set MMM defaults or user defined overrides
        chilife.set_lj_params(user_lj)

        clash_radius = kwargs.pop("clash_radius", MMM_maxdist[label] + 4)
        alignment_method = kwargs.pop("alignment_method", "mmm")
        clash_ori = kwargs.pop("clash_ori", "CA")
        energy_func = kwargs.pop(
            "energy_func", partial(chilife.get_lj_energy, cap=np.inf)
        )
        use_H = kwargs.pop("use_H", True)
        forgive = kwargs.pop("forgive", 0.5)

        # Calculate the SpinLabel
        SL = chilife.SpinLabel(
            label,
            site,
            protein,
            chain,
            alignment_method=alignment_method,
            clash_radius=clash_radius,
            clash_ori=clash_ori,
            energy_func=energy_func,
            use_H=use_H,
            forgive=forgive,
            **kwargs,
        )

        # restore the force field parameter set being used before creating the spin label
        chilife.set_lj_params(curr_lj)
        return SL

    @classmethod
    def from_wizard(
        cls,
        label,
        site=1,
        protein=None,
        chain=None,
        to_find=200,
        to_try=10000,
        vdw=2.5,
        clashes=5,
        **kwargs,
    ):
        """Create a SpinLabel object using the default MTSSLWizard protocol with any modifications passed via kwargs"""

        prelib = cls(label, site, protein, chain, eval_clash=False, **kwargs)
        prelib.sigmas = np.array([])

        coords = np.zeros((to_find, len(prelib.atom_names), 3))
        internal_coords = []
        i = 0
        if protein is not None:
            protein_clash_idx = prelib.protein_tree.query_ball_point(
                prelib.centroid(), 19.0
            )
            protein_clash_idx = [
                idx for idx in protein_clash_idx if idx not in prelib.clash_ignore_idx
            ]

        a, b = [list(x) for x in zip(*prelib.non_bonded)]
        for _ in range(np.rint(to_try / to_find).astype(int)):
            sample, _, internal_sample = prelib.sample(
                n=to_find, off_rotamer=True, return_dihedrals=True
            )

            # Evaluate internal clashes
            dist = np.linalg.norm(sample[:, a] - sample[:, b], axis=2)
            sidx = np.atleast_1d(np.squeeze(np.argwhere(np.all(dist > 2, axis=1))))
            if len(sidx) == 0:
                continue

            sample = sample[sidx, ...]
            internal_sample = internal_sample[sidx]
            if protein is not None:
                # Evaluate external clashes
                dist = cdist(
                    sample[:, prelib.side_chain_idx].reshape(-1, 3),
                    prelib.protein_tree.data[protein_clash_idx],
                )
                dist = dist.reshape(len(sidx), -1)
                nclashes = np.sum(dist < vdw, axis=1)
                sidx = np.atleast_1d(np.squeeze(np.argwhere(nclashes < clashes)))
            else:
                sidx = np.arange(len(sample))

            if len(sidx) == 0:
                continue

            if i + len(sidx) >= to_find:
                sidx = sidx[: to_find - i]

            coords[i: i + len(sidx)] = sample[sidx]
            internal_coords.append(internal_sample[sidx])
            i += len(sidx)

            if i == to_find:
                break

        coords = coords[coords.sum(axis=(1, 2)) != 0]
        prelib.internal_coords = (
            np.concatenate(internal_coords) if len(internal_coords) > 0 else []
        )
        prelib._dihedrals = np.rad2deg(
            [IC.get_dihedral(1, prelib.dihedral_atoms) for IC in prelib.internal_coords]
        )
        prelib._coords = coords
        prelib.weights = np.ones(len(coords))
        prelib.weights /= prelib.weights.sum()
        return prelib


    def _base_copy(self, rotlib=None):
        return chilife.SpinLabel(self.res, self.site, rotlib=rotlib, chain=self.chain)



import os, urllib, pickle, math, rtoml
from operator import itemgetter
from pathlib import Path
from typing import Set, List, Union, Tuple
from numpy.typing import ArrayLike
from dataclasses import dataclass
from collections import Counter, defaultdict
import MDAnalysis
import numpy as np
from scipy.spatial import cKDTree
from scipy.spatial.distance import cdist

from MDAnalysis.core.topologyattrs import Atomindices, Resindices, Segindices, Segids
import MDAnalysis as mda

import chilife
from .numba_utils import get_sasa
from .RotamerEnsemble import RotamerEnsemble
from .dRotamerEnsemble import dRotamerEnsemble

import networkx as nx


def get_dihedral_rotation_matrix(theta: float, v: ArrayLike) -> ArrayLike:
    """Build a matrix that will rotate coordinates about a vector, v, by theta in radians.

    Parameters
    ----------
    theta : float
        Rotation angle in radians.
    v : (3,) ArrayLike
        Three dimensional vector to rotate about.

    Returns
    -------
    rotation_matrix : np.ndarray
            Matrix that will rotate coordinates about the vector, V by angle theta.
    """

    # Normalize input vector
    v = v / np.linalg.norm(v)

    # Compute Vx matrix
    Vx = np.zeros((3, 3))
    Vx[[2, 0, 1], [1, 2, 0]] = v
    Vx -= Vx.T

    # Rotation matrix. See https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    rotation_matrix = (
            np.identity(3) * np.cos(theta)
            + np.sin(theta) * Vx
            + (1 - np.cos(theta)) * np.outer(v, v)
    )

    return rotation_matrix


def get_dihedral(p: ArrayLike) -> float:
    """Calculates dihedral of a given set of atoms, ``p`` . Returns value in radians.

     .. code-block:: python

                         3
              ------>  /
             1-------2
           /
         0

    Parameters
    ----------
    p : (4, 3) ArrayLike
        Matrix containing coordinates to be used to calculate dihedral.

    Returns
    -------
    dihedral : float
        Dihedral angle in radians.
    """

    # Unpack p
    p0, p1, p2, p3 = p

    # Define vectors from coordinates
    b0 = -1.0 * (p1 - p0)
    b1 = p2 - p1
    b2 = p3 - p2

    # Normalize dihedral bond vector
    b1 /= np.linalg.norm(b1)

    # Calculate dihedral projections orthogonal to the bond vector
    v = b0 - np.dot(b0, b1) * b1
    w = b2 - np.dot(b2, b1) * b1

    # Calculate angle between projections
    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)

    dihedral = math.atan2(y, x)

    return dihedral


def get_angle(p: ArrayLike) -> float:
    r"""Calculate the angle created by 3 points.

     .. code-block:: python

               p2
             / θ \
          p1      p3

    Parameters
    ----------
    p: ArrayLike :
        Array of three points to calculate the angle between.

    Returns
    -------
    angle : float
        Angle created by the three points.

    """
    p1, p2, p3 = p
    v1 = p1 - p2
    v2 = p3 - p2
    X = v1 @ v2
    Y = np.cross(v1, v2)
    Y = math.sqrt(Y @ Y)

    angle = math.atan2(Y, X)

    return angle


def set_dihedral(p: ArrayLike, angle: float, mobile: ArrayLike) -> ArrayLike:
    """Sets the dihedral angle by rotating all ``mobile`` atoms from their current position about the dihedral bond
    defined by the four atoms in ``p`` . Dihedral will be set to the value of ``angle`` in degrees.

    Parameters
    ----------
    p : ArrayLike
        Coordinates of atoms that define dihedral to rotate about.
    angle : float
        New angle to set the dihedral to (degrees).
    mobile : np.ndarray
        Atom coordinates to move by setting dihedral.

    Returns
    -------
    new_mobile : np.ndarray
        New positions for the mobile atoms
    """

    current = get_dihedral(p)
    angle = np.deg2rad(angle) - current
    angle = angle

    ori = p[1]
    mobile -= ori
    v = p[2] - p[1]
    v /= np.linalg.norm(v)
    R = get_dihedral_rotation_matrix(angle, v)

    new_mobile = R.dot(mobile.T).T + ori

    return new_mobile


def local_mx(*p, method: Union[str, callable] = "bisect") -> Tuple[ArrayLike, ArrayLike]:
    """Calculates a translation vector and rotation matrix to transform a set of coordinates from the global
    coordinate frame to a local coordinate frame defined by ``p`` , using the specified method.

    Parameters
    ----------
    p : ArrayLike
        3D coordinates of the three points defining the coordinate system (Usually N, CA, C).
    method : str, callable
        Method to use for generation of rotation matrix

    Returns
    -------
    origin : np.ndarray
        Cartesian coordinate of the origin to be subtracted from the coordinates before applying the rotation matrix.
    rotation_matrix : np.ndarray
        Rotation matrix to transform a set of coordinates to the local frame defined by p and the selected method.
    """

    if isinstance(method, str):
        method = chilife.alignment_methods[method]

    p1, p2, p3 = p

    if method.__name__ == 'fit_alignment':
        rotation_matrix, _ = method(p1, p2, p3)
        origin = np.mean([p1[0], p2[0], p3[0]], axis=0)
    else:
        # Transform coordinates such that the CA atom is at the origin
        p1n = p1 - p2
        p3n = p3 - p2
        p2n = p2 - p2

        origin = p2

        # Local Rotation matrix is the inverse of the global rotation matrix
        rotation_matrix, _ = method(p1n, p2n, p3n)

    rotation_matrix = rotation_matrix.T

    return origin, rotation_matrix


def global_mx(*p: ArrayLike, method: Union[str, callable] = "bisect") -> Tuple[ArrayLike, ArrayLike]:
    """Calculates a translation vector and rotation matrix to transform the a set of coordinates from the local
    coordinate frame to the global coordinate frame using the specified method.

    Parameters
    ----------
    p : ArrayLike
        3D coordinates of the three points used to define the new coordinate system (Usually N, CA, C)
    method : str
        Method to use for generation of rotation matrix

    Returns
    -------
    rotation_matrix : np.ndarray
        Rotation matrix to be applied to the set of coordinates before translating
    origin : np.ndarray
        Vector to be added to the coordinates after rotation to translate the coordinates to the global frame.
    """

    if isinstance(method, str):
        method = chilife.alignment_methods[method]

    if method.__name__ == 'fit_alignment':
        p = [pi[::-1] for pi in p]

    rotation_matrix, origin = method(*p)
    return rotation_matrix, origin


@dataclass
class FreeAtom:
    """Atom class for atoms in cartesian space.

    Attributes
    ----------
    name : str
        Atom name
    atype : str
        Atom type
    index : int
        Atom number
    resn : str
        Name of the residue that the atom belongs to
    resi : int
        The residue index/number that the atom belongs to
    coords : np.ndarray
        The cartesian coordinates of the Atom
    """

    name: str
    atype: str
    index: int
    resn: str
    resi: int
    coords: np.ndarray


def save_ensemble(name: str, atoms: ArrayLike, coords: ArrayLike = None) -> None:
    """Save a rotamer ensemble as multiple states of the same molecule.

    Parameters
    ----------
    name : str
        file name to save rotamer ensemble to
    atoms : ArrayLike
        list of Atom objects
    coords : ArrayLike
        Array of atom coordinates corresponding to Atom objects
    """

    if not name.endswith(".pdb"):
        name += ".pdb"

    if coords is None and isinstance(atoms[0], list):
        with open(name, "w", newline="\n") as f:
            for i, model in enumerate(atoms):
                f.write(f"MODEL {i + 1}\n")
                for atom in model:
                    f.write(
                        f"ATOM  {atom.index + 1:5d}  {atom.name:<4s}{atom.resn:3s} {'A':1s}{atom.resi:4d}   "
                        f"{atom._coords[0]:8.3f}{atom._coords[1]:8.3f}{atom._coords[2]:8.3f}{1.0:6.2f}{1.0:6.2f}        "
                        f"  {atom.atype:>2s}\n"
                    )
                f.write("ENDMDL\n")

    elif len(coords.shape) > 2:
        with open(name, "w", newline="\n") as f:
            for i, model in enumerate(coords):
                f.write(f"MODEL {i + 1}\n")
                for atom, coord in zip(atoms, model):
                    f.write(
                        f"ATOM  {atom.index + 1:5d}  {atom.name:<4s}{atom.resn:3s} {'A':1s}{atom.resi:4d}   "
                        f"{coord[0]:8.3f}{coord[1]:8.3f}{coord[2]:8.3f}{1.0:6.2f}{1.0:6.2f}          {atom.atype:>2s}\n"
                    )
                f.write("ENDMDL\n")

    else:
        save_pdb(name, atoms, coords)


def save_pdb(name: Union[str, Path], atoms: ArrayLike, coords: ArrayLike, mode: str = "w") -> None:
    """Save a single state pdb structure of the provided atoms and coords.

    Parameters
    ----------
    name : str, Path
        Name or Path object of file to save
    atoms : ArrayLike
        List of Atom objects to be saved
    coords : ArrayLike
        Array of atom coordinates corresponding to atoms
    mode : str
        File open mode. Usually used to specify append ("a") when you want to add structures to a PDB rather than
        overwrite that pdb.
    """
    name = Path(name) if isinstance(name, str) else name
    name = name.with_suffix(".pdb")

    with open(name, mode, newline="\n") as f:
        f.write('MODEL\n')
        for atom, coord in zip(atoms, coords):
            f.write(
                f"ATOM  {atom.index + 1:5d} {atom.name:^4s} {atom.resn:3s} {'A':1s}{atom.resi:4d}    "
                f"{coord[0]:8.3f}{coord[1]:8.3f}{coord[2]:8.3f}{1.0:6.2f}{1.0:6.2f}          {atom.atype:>2s}  \n"
            )
        f.write('ENDMDL\n')

def write_atoms(f, atoms: ArrayLike, coords: ArrayLike) -> None:
    """Save a single state pdb structure of the provided atoms and coords.

    Parameters
    ----------
    f : IO object

    atoms : ArrayLike
        List of Atom objects to be saved
    coords : ArrayLike
        Array of atom coordinates corresponding to atoms
    """

    for atom, coord in zip(atoms, coords):
        f.write(
            f"ATOM  {atom.index + 1:5d} {atom.name:^4s} {atom.resn:3s} {'A':1s}{atom.resi:4d}    "
            f"{coord[0]:8.3f}{coord[1]:8.3f}{coord[2]:8.3f}{1.0:6.2f}{1.0:6.2f}          {atom.atype:>2s}  \n"
        )


def get_missing_residues(
        protein: Union[MDAnalysis.Universe, MDAnalysis.AtomGroup],
        ignore: Set[int] = None,
        use_H: bool = False,
) -> List:
    """Get a list of RotamerEnsemble objects corresponding to the residues of the provided protein that are missing
    heavy atoms.

    Parameters
    ----------
    protein : MDAnalysis.Universe, MDAnalysis.AtomGroup
        Protein to search for residues with missing atoms.
    ignore : set
        List of residue numbers to ignore. Usually sites you plan to label or mutate.
    use_H : bool
        Whether the new side chain should have hydrogen atoms.

    Returns
    -------
    missing_residues : list
        A list of RotamerEnsemble objects corresponding to residues with missing heavy atoms.
    """
    ignore = set() if ignore is None else ignore
    missing_residues = []
    cache = {}

    for res in protein.residues:
        # Only consider supported residues because otherwise chiLife wouldn't know what's missing
        if (
                res.resname not in chilife.SUPPORTED_RESIDUES
                or res.resnum in ignore
                or res.resname in ["ALA", "GLY"]
        ):
            continue

        # Check if there are any missing heavy atoms
        heavy_atoms = res.atoms.types[res.atoms.types != "H"]
        if len(heavy_atoms) != cache.get(res.resname, len(RotamerEnsemble(res.resname).atom_names)):
            missing_residues.append(
                RotamerEnsemble(
                    res.resname,
                    res.resnum,
                    protein=protein,
                    chain=res.segid,
                    use_H=use_H,
                )
            )

    return missing_residues


def mutate(
        protein: MDAnalysis.Universe,
        *ensembles: RotamerEnsemble,
        add_missing_atoms: bool = True,
        random_rotamers: bool = False,
) -> MDAnalysis.Universe:
    """Create a new Universe where the native residue is replaced with the highest probability rotamer from a
    RotamerEnsemble or SpinLabel object.

    Parameters
    ----------
    protein : MDAnalysis.Universe
        An MDA Universe object containing protein to be spin labeled
    ensembles : RotamerEnsemble, SpinLabel
        Precomputed RotamerEnsemble or SpinLabel object to use for selecting and replacing the spin native amino acid
    random_rotamers :bool
        Randomize rotamer conformations
    add_missing_atoms : bool
        Model side chains missing atoms if they are not present in the provided structure.

    Returns
    -------
    U : MDAnalysis.Universe
        New Universe with a copy of the spin labeled protein with the highest probability rotamer
    """

    # Check for dRotamerEnsembles in ensembles
    temp_ensemble = []
    for lib in ensembles:
        if isinstance(lib, (RotamerEnsemble, dRotamerEnsemble)):
            temp_ensemble.append(lib)
        else:
            raise TypeError(
                f"mutate only accepts RotamerEnsemble, SpinLabel and dSpinLabel objects, not {lib}."
            )

    ensembles = temp_ensemble

    if add_missing_atoms:
        if len(ensembles) > 0 and all(not hasattr(lib, "H_mask") for lib in ensembles):
            use_H = True
        elif any(not hasattr(lib, "H_mask") for lib in ensembles):
            raise AttributeError(
                "User provided some ensembles with hydrogen atoms and some without. Make sure all "
                "ensembles either do or do not use hydrogen"
            )
        else:
            use_H = False

        missing_residues = get_missing_residues(
            protein, ignore={res.site for res in ensembles}, use_H=use_H
        )
        ensembles = list(ensembles) + missing_residues

    label_sites = {}
    for spin_label in ensembles:
        if isinstance(spin_label, dRotamerEnsemble):
            label_sites[spin_label.site1, spin_label.chain] = spin_label
            label_sites[spin_label.site2, spin_label.chain] = spin_label
        else:
            label_sites[spin_label.site, spin_label.chain] = spin_label

    protein = protein.select_atoms(
        f'(not altloc B) and (not (byres name OH2 or resname HOH))'
    )
    label_selstr = " or ".join([f"({label.selstr})" for label in ensembles])
    other_atoms = protein.select_atoms(f"not ({label_selstr})")

    resids = [res.resid for res in protein.residues]

    # Allocate lists for universe information
    atom_info = []
    res_names = []
    segidx = []

    # Loop over residues in old universe
    for i, res in enumerate(protein.residues):
        resloc = (res.resnum, res.segid)

        # If the residue is the spin labeled residue replace it with the highest probability spin label
        if resloc in label_sites:
            rot_ens = label_sites[resloc]
            if isinstance(rot_ens, dRotamerEnsemble):
                r1l = len(rot_ens.rl1mask)
                r2l = len(rot_ens.rl2mask)
                both = r1l + r2l

                if resloc[0] == rot_ens.site1:
                    # Add site 1
                    atom_info += [
                        (i, name, atype)
                        for name, atype in zip(rot_ens.atom_names[:r1l], rot_ens.atom_types[:r2l])
                    ]

                elif resloc[0] == rot_ens.site2:
                    atom_info += [
                        (i, name, atype)
                        for name, atype in zip(rot_ens.atom_names[r1l:r1l + r2l], rot_ens.atom_types[r1l:r1l + r2l])
                    ]
                    # Add cap
                    atom_info += [
                        (i, name, atype)
                        for name, atype in zip(rot_ens.atom_names[both:], rot_ens.atom_types[both:])
                    ]
                else:
                    raise RuntimeError("The residue specified is not part of the dRotamerEnsemble being constructed")
            else:
                atom_info += [
                    (i, name, atype)
                    for name, atype in zip(rot_ens.atom_names, rot_ens.atom_types)
                ]

            # Add missing Oxygen from rotamer ensemble
            res_names.append(rot_ens.res)
            segidx.append(rot_ens.segindex)

        # Else retain the atom information from the parent universe
        else:
            atom_info += [
                (i, atom.name, atom.type) for atom in res.atoms if atom.altLoc != "B"
            ]
            res_names.append(res.resname)
            segidx.append(res.segindex)

    # Reindex segments in case any were dropped from the parent universe
    idxmap = {idx: i for i, idx in enumerate(np.unique(segidx))}
    segidx = np.fromiter((idxmap[idx] for idx in segidx), dtype=int)

    # Unzip atom information into individual lists
    residx, atom_names, atom_types = zip(*atom_info)
    segids = list(Counter(protein.residues.segids))
    # Allocate a new universe with the appropriate information

    if isinstance(protein, (mda.Universe, mda.AtomGroup)):
        U = make_mda_uni(atom_names, atom_types, res_names, residx, resids, segidx, segids)
    elif isinstance(protein, chilife.MolecularSystem):
        U = chilife.Protein.from_arrays(atom_names, atom_types, res_names, residx, resids, segidx, segids)

    # Apply old coordinates to non-spinlabel atoms
    new_other_atoms = U.select_atoms(f"not ({label_selstr})")
    new_other_atoms.atoms.positions = other_atoms.atoms.positions

    # Apply most probable spin label coordinates to spin label atoms
    for spin_label in label_sites.values():
        sl_atoms = U.select_atoms(spin_label.selstr)
        if random_rotamers:
            sl_atoms.atoms.positions = spin_label.coords[
                np.random.choice(len(spin_label.coords), p=spin_label.weights)
            ]
        else:
            sl_atoms.atoms.positions = spin_label.coords[np.argmax(spin_label.weights)]

    return U


def randomize_rotamers(
        protein: Union[mda.Universe, mda.AtomGroup],
        rotamer_libraries: List[RotamerEnsemble],
        **kwargs,
) -> None:
    """Modify a protein object in place to randomize side chain conformations.

    Parameters
    ----------
    protein : MDAnalysis.Universe, MDAnalysis.AtomGroup
        Protein object to modify.
    rotamer_libraries : list
        RotamerEnsemble objects attached to the protein corresponding to the residues to be repacked/randomized.
    **kwargs : dict
        Additional Arguments to pass to ``sample`` method. See :mod:`sample <chiLife.RotamerEnsemble.sample>` .
    """
    for rotamer in rotamer_libraries:
        coords, weight = rotamer.sample(off_rotamer=kwargs.get("off_rotamer", False))
        mask = ~np.isin(protein.ix, rotamer.clash_ignore_idx)
        protein.atoms[~mask].positions = coords


def get_sas_res(
        protein: Union[mda.Universe, mda.AtomGroup], cutoff: float = 30
) -> Set[Tuple[int, str]]:
    """Run FreeSASA to get solvent accessible surface residues in the provided protein

    Parameters
    ----------
    protein : MDAnalysis.Universe, MDAnalysis.AtomGroup
        Protein object to measure Solvent Accessible Surfaces (SAS) area of and report the SAS residues.
    cutoff : float
        Exclude residues from list with SASA below cutoff in angstroms squared.

    Returns
    -------
    SAResi : set
        Set of solvent accessible surface residues.

    """
    environment_coords = protein.atoms.positions
    environment_radii = chilife.get_lj_rmin(protein.atoms.types)
    atom_sasa = get_sasa(environment_coords, environment_radii, by_atom=True)

    SASAs = {(residue.resnum, residue.segid) for residue in protein.residues if
             atom_sasa[0, residue.atoms.ix].sum() >= cutoff}

    return SASAs


def fetch(accession_number: str, save: bool = False) -> MDAnalysis.Universe:
    """Fetch pdb file from the protein data bank or the AlphaFold Database and optionally save to disk.

    Parameters
    ----------
    accession_number : str
        4 letter structure PDB ID or alpha fold accession number. Note that AlphaFold accession numbers must begin with
        'AF-'.
    save : bool
        If true the fetched PDB will be saved to the disk.

    Returns
    -------
    U : MDAnalysis.Universe
        MDAnalysis Universe object of the protein corresponding to the provided PDB ID or AlphaFold accession number

    """
    accession_number = accession_number.split('.pdb')[0]
    pdb_name = accession_number + '.pdb'

    if accession_number.startswith('AF-'):
        print(f"https://alphafold.ebi.ac.uk/files/{accession_number}-F1-model_v3.pdb")
        urllib.request.urlretrieve(f"https://alphafold.ebi.ac.uk/files/{accession_number}-F1-model_v3.pdb", pdb_name)
    else:
        urllib.request.urlretrieve(f"http://files.rcsb.org/download/{pdb_name}", pdb_name)

    U = mda.Universe(pdb_name, in_memory=True)

    if not save:
        os.remove(pdb_name)

    return U


def atom_sort_key(pdb_line: str) -> Tuple[str, int, int]:
    """Assign a base rank to sort atoms of a pdb.

    Parameters
    ----------
    pdb_line : str
        ATOM line from a pdb file as a string.

    Returns
    -------
    tuple :
        chain_id, resid, name_order.
        ordered ranking of atom for sorting the pdb.
    """
    chain_id = pdb_line[21]
    res_name = pdb_line[17:20].strip()
    resid = int(pdb_line[22:26].strip())
    atom_name = pdb_line[12:17].strip()
    atom_type = pdb_line[76:79].strip()
    if res_name == "ACE":
        if atom_type != 'H' and atom_name not in ('CH3', 'C', 'O'):
            raise ValueError(f'"{atom_name}" is not canonical name of an ACE residue atom. \n'
                             f'Please rename to "CH3", "C", or "O"')
        name_order = (
            {"CH3": 0, "C": 1, "O": 2}.get(atom_name, 4) if atom_type != "H" else 5
        )

    else:
        name_order = atom_order.get(atom_name, 4) if atom_type != "H" else atom_order.get(atom_name, 7)

    return chain_id, resid, name_order


def pose2mda(pose) -> MDAnalysis.Universe:
    """Create an MDAnalysis universe from a pyrosetta pose

    Parameters
    ----------
    pose : pyrosetta.rosetta.core.Pose
        pyrosetta pose object.

    Returns
    -------
    mda_protein : MDAnalysis.Universe
        Copy of the input pose as an MDAnalysis Universe object
    """
    coords = np.array(
        [
            res.xyz(atom)
            for res in pose.residues
            for atom in range(1, res.natoms() + 1)
            if res.atom_type(atom).element().strip() != "X"
        ]
    )
    atypes = np.array(
        [
            str(res.atom_type(atom).element()).strip()
            for res in pose.residues
            for atom in range(1, res.natoms() + 1)
            if res.atom_type(atom).element().strip() != "X"
        ]
    )
    anames = np.array(
        [
            str(res.atom_name(atom)).strip()
            for res in pose.residues
            for atom in range(1, res.natoms() + 1)
            if res.atom_type(atom).element().strip() != "X"
        ]
    )
    resindices = np.array(
        [
            res.seqpos() - 1
            for res in pose.residues
            for atom in range(1, res.natoms() + 1)
            if res.atom_type(atom).element().strip() != "X"
        ]
    )

    n_residues = len(pose)

    segindices = np.array([0] * n_residues)
    resnames = np.array([res.name() for res in pose])
    resnums = np.array([res.seqpos() for res in pose])

    mda_protein = make_mda_uni(anames, atypes, resnames, resindices, resnums, segindices)
    mda_protein.positions = coords

    return mda_protein


def guess_bonds(coords: ArrayLike, atom_types: ArrayLike) -> np.ndarray:
    """ Given a set of coordinates and their atom types (elements) guess the bonds based off an empirical metric.

    Parameters
    ----------
    coords : ArrayLike
        Array of three-dimensional coordinates of the atoms of a molecule or set of molecules for which you would like
        to guess the bonds of.
    atom_types : ArrayLike
        Array of the element symbols corresponding to the atoms of ``coords``

    Returns
    -------
    bonds : np.ndarray
        An array of the atom index pairs corresponding to the atom pairs that are thought ot form bonds.
    """
    kdtree = cKDTree(coords)
    pairs = kdtree.query_pairs(4., output_type='ndarray')
    pair_names = [tuple(x) for x in atom_types[pairs].tolist()]
    bond_lengths = itemgetter(*pair_names)(chilife.bond_hmax_dict)
    a_atoms = pairs[:, 0]
    b_atoms = pairs[:, 1]

    dist = np.linalg.norm(coords[a_atoms] - coords[b_atoms], axis=1)
    bonds = pairs[dist < bond_lengths]

    return bonds


def get_min_topol(lines: List[List[str]],
                  forced_bonds: set = None) -> Set[Tuple[int, int]]:
    """ Git the minimum topology shared by all the states/models a PDB ensemble. This is to ensure a consistent
    internal coordinate system between all conformers of an ensemble even when there are minor differences in topology.
    e.g. when dHis-Cu-NTA has the capping ligand in different bond orientations.

    Parameters
    ----------
    lines : List[List[str]]
        List of lists corresponding to individual states/models of a pdb file. All models must have the same stoma in
        the same order and only the coordinates should differ.
    forced_bonds : set
        A set of bonds to that must be used regardless even if the bond lengths are not physically reasonable.
    Returns
    -------
    minimal_bond_set : Set[Tuple[int, int]]
        A set of tuples holding the indices of atom pairs which are thought to be bonded in all states/models.
    """
    bonds_list = []
    if isinstance(lines[0], str):
        lines = [lines]

    # Get bonds for all structures
    for struct in lines:
        coords = np.array([(line[30:38], line[38:46], line[46:54]) for line in struct], dtype=float)
        atypes = np.array([line[76:78].strip() for line in struct])
        pairs = guess_bonds(coords, atypes)
        bonds = set(tuple(pair) for pair in pairs)
        bonds_list.append(bonds)

    # Get the shared bonds between all structures.
    minimal_bond_set = set.intersection(*bonds_list)
    # Include any forced bonds
    if forced_bonds is not None:
        minimal_bond_set |= forced_bonds

    return minimal_bond_set


def parse_connect(connect: List[str]) -> Tuple[Set[Tuple[int]]]:
    """
    Parse PDB CONECT information to get a list covalent bonds, hydrogen bonds and ionic bonds.

    Parameters
    ----------
    connect : List[str]
        A list of strings that are the CONECT lines from a PDB file.

    Returns
    -------
    c_bonds : Set[Tuple[int]]
        Set of atom index pairs corresponding to atoms that are bound covalently.
    h_bonds : Set[Tuple[int]]
        Set of atom index pairs corresponding to atoms that are hydrogen bound.
    ci_bonds : Set[Tuple[int]]
        Set of atom index pairs corresponding to atoms that are bound ionically.
    """
    c_bonds, h_bonds, i_bonds = set(), set(), set()
    for line in connect:
        line = line.ljust(61)
        a0 = int(line[6:11])
        c_bonds |= {tuple(sorted((a0 - 1, int(b) - 1))) for b in line[11:31].split()}
        h_bonds |= {tuple(sorted((a0 - 1, int(b) - 1))) for b in (line[31:41].split() + line[46:56].split())}
        i_bonds |= {tuple(sorted((a0 - 1, int(b) - 1))) for b in (line[41:46], line[56:61]) if not b.isspace()}

    return c_bonds, h_bonds, i_bonds


def sort_pdb(pdbfile: Union[str, List[str], List[List[str]]],
             uniform_topology: bool = True,
             index: bool = False,
             bonds: Union[ArrayLike, Set] = set(),
             **kwargs) -> Union[List[str], List[List[str]], List[int]]:
    """Read ATOM lines of a pdb and sort the atoms according to chain, residue index, backbone atoms and side chain atoms.
    Side chain atoms are sorted by distance to each other/backbone atoms with atoms closest to the backbone coming
    first and atoms furthest from the backbone coming last. This sorting is essential to making internal-coordinates
    with consistent and preferred dihedral definitions.

    Parameters
    ----------
    pdbfile : str, List[str], List[List[str]]
        Name of the PDB file, a list of strings containing ATOM lines of a PDB file or a list of lists containing
        ATOM lines of a PDB file, where each sublist corresponds to a state/model of a multi-state pdb.
    uniform_topology: bool
        When given a multi-state pdb, assume that all states have the same topology (bonds) as the first state.
    index: bool :
         Return the sorted index rather than the sorted lines.
    bonds: ArrayLike :
         When sorting the PDB, use the provided bond list to as the topology rather than guessing the bonds.

    Returns
    -------
    lines : List[str], List[List[str]]
        Sorted list of strings corresponding to the ATOM entries of a PDB file.
    """

    if isinstance(pdbfile, (str, Path)):
        with open(pdbfile, "r") as f:
            lines = f.readlines()

        start_idxs = []
        end_idxs = []
        connect = []
        lines = [line for line in lines if line.startswith(('MODEL', 'ENDMDL', 'CONECT', 'ATOM', 'HETATM'))]
        for i, line in enumerate(lines):
            if line.startswith('MODEL'):
                start_idxs.append(i + 1)
            elif line.startswith("ENDMDL"):
                end_idxs.append(i)
            elif line.startswith('CONECT'):
                connect.append(line)

        # Use connect information for bonds if present
        if connect != [] and bonds == set():
            bonds, _, _ = parse_connect(connect)

        # If it's a multi-state pdb...
        if start_idxs != []:

            if uniform_topology:
                # Assume that all states have the same topology as the first
                idxs = _sort_pdb_lines(lines[start_idxs[0]:end_idxs[0]], bonds=bonds, index=True, **kwargs)

            else:
                # Calculate the shared topology and force it
                min_bonds_list = get_min_topol(lines[start_idxs[0]:end_idxs[0]], forced_bonds=bonds)
                idxs = _sort_pdb_lines(lines[start_idxs[0]:end_idxs[0]], bonds=min_bonds_list, index=True, **kwargs)

            if isinstance(idxs, tuple):
                idxs, bonds = idxs

            lines[:] = [[lines[idx + start][:6] + f"{i + 1:5d}" + lines[idx + start][11:]
                         for i, idx in enumerate(idxs)]
                        for start in start_idxs]

            if kwargs.get('return_bonds', False):
                lines = lines, bonds
        else:
            lines = _sort_pdb_lines(lines, bonds=bonds, index=index, **kwargs)

    elif isinstance(pdbfile, list):
        lines = _sort_pdb_lines(pdbfile, bonds=bonds, index=index, **kwargs)

    return lines


def _sort_pdb_lines(lines, bonds=None, index=False, **kwargs) -> \
        Union[List[str], List[int], Tuple[list[str], List[Tuple[int]]]]:
    """
    Helper function to sort PDB ATOM and HETATM lines based off of the topology of the topology of the molecule.

    Parameters
    ----------
    lines : List[str]
        A list of the PDB ATOM and HETATM lines.
    bonds : Set[Tuple[int]]
        A Set of tuples of atom indices corresponding to atoms ( `lines` ) that are bound to each other.
    index : bool
        If True a list of atom indices will be returned
    **kwargs : dict
        Additional keyword arguments.
        return_bonds : bool
            Return bond indices as well, usually only used when letting the function guess the bonds.

    Returns
    -------
    lines : List[str | int]
        The sorted lines or indices corresponding to the sorted lines.
    bonds: Set[Tuple[int]]
        A set of tuples containing pars of indices corresponding to the atoms bound to in lines.
    """

    lines = [line for line in lines if line.startswith(("ATOM", "HETATM"))]
    n_atoms = len(lines)
    index_key = {line[6:11]: i for i, line in enumerate(lines)}

    # Presort
    lines.sort(key=atom_sort_key)
    presort_idx_key = {line[6:11]: i for i, line in enumerate(lines)}
    presort_bond_key = {index_key[line[6:11]]: i for i, line in enumerate(lines)}

    coords = np.array([[float(line[30:38]), float(line[38:46]), float(line[46:54])] for line in lines])
    atypes = np.array([line[76:78].strip() for line in lines])
    anames = np.array([line[12:17].strip() for line in lines])

    if bonds:
        input_bonds = {tuple(b) for b in bonds}
        presort_bonds = set(tuple(sorted((presort_bond_key[b1], presort_bond_key[b2]))) for b1, b2 in bonds)
    else:
        bonds = guess_bonds(coords, atypes)
        presort_bonds = set(tuple(sorted((b1, b2))) for b1, b2 in bonds)

    # get residue groups
    chain, resi = lines[0][21], int(lines[0][22:26].strip())
    start = 0
    resdict = {}
    for curr, pdb_line in enumerate(lines):

        if chain != pdb_line[21] or resi != int(pdb_line[22:26].strip()):
            resdict[chain, resi] = start, curr
            start = curr
            chain, resi = pdb_line[21], int(pdb_line[22:26].strip())

    resdict[chain, resi] = start, curr + 1
    midsort_key = []
    for key in resdict:
        start, stop = resdict[key]
        n_heavy = np.sum(atypes[start:stop] != 'H')

        #  Force N, CA, C,
        if np.array_equal(anames[start: start + 4], ['N', 'CA', 'C', 'O']):
            sorted_args = [0, 1, 2, 3]
        # if not a canonical and the first amino acid use the first heavy atom
        elif start == 0:
            sorted_args = [0]
        else:
            # If not a connected via peptide backbone
            for a, b in presort_bonds:
                # Find the atom bonded to a previous residue
                if a < start and start <= b < stop and atypes[b] != 'H':
                    sorted_args = [b - start]
                    break
                # Otherwise get the closest to any previous atom
            else:
                dist_mat = cdist(coords[:start], coords[start:stop])
                sorted_args = [np.squeeze(np.argwhere(dist_mat == dist_mat.min()))[1]]

        if len(sorted_args) != n_heavy:

            root_idx = 1 if len(sorted_args) == 4 else sorted_args[0]
            bonds = np.array([bond for bond in presort_bonds
                              if (start <= bond[0] < stop) and (start <= bond[1] < stop)])
            bonds -= start
            bonds = np.asarray(bonds)

            # Get all nearest neighbors and sort by distance
            try:
                distances = np.linalg.norm(coords[start:stop][bonds[:, 0]] - coords[start:stop][bonds[:, 1]], axis=1)
            except:
                breakpoint()
            distances = np.around(distances, decimals=3)
            idx_sort = np.lexsort((bonds[:, 0], bonds[:, 1], distances))
            pairs = bonds[idx_sort]
            pairs = [pair for pair in pairs if np.any(~np.isin(pair, sorted_args))]

            graph = nx.Graph()
            graph.add_edges_from(pairs)

            # Start stemming from CA atom
            CA_edges = [edge[1] for edge in nx.bfs_edges(graph, root_idx) if edge[1] not in sorted_args]

            # Check for disconnected parts of residue
            if not nx.is_connected(graph):
                for g in nx.connected_components(graph):
                    if np.any([arg in g for arg in sorted_args]):
                        continue
                    CA_nodes = [idx for idx in CA_edges if atypes[start + idx] != 'H']
                    g_nodes = [idx for idx in g if atypes[start + idx] != 'H']
                    near_root = cdist(coords[start:stop][CA_nodes], coords[start:stop][g_nodes]).argmin()

                    yidx = near_root % len(g_nodes)
                    CA_edges += [g_nodes[yidx]] + [edge[1]
                                                   for edge in nx.bfs_edges(graph, g_nodes[yidx])
                                                   if edge[1] not in sorted_args]
        elif stop - start > n_heavy:
            # Assumes  non-heavy atoms come after the heavy atoms, which should be true because of the pre-sort
            CA_edges = list(range(n_heavy, n_heavy + (stop - start - len(sorted_args))))
        else:
            CA_edges = []

        sorted_args = sorted_args + CA_edges

        # get any leftover atoms (eg HN)
        if len(sorted_args) != stop - start:
            for idx in range(stop - start):
                if idx not in sorted_args:
                    sorted_args.append(idx)

        midsort_key += [x + start for x in sorted_args]

    lines[:] = [lines[i] for i in midsort_key]
    lines.sort(key=atom_sort_key)

    if 'input_bonds' not in locals():
        input_bonds = presort_bonds
        idxmap = {presort_idx_key[line[6:11]]: i for i, line in enumerate(lines)}
    else:
        idxmap = {index_key[line[6:11]]: i for i, line in enumerate(lines)}

    # Return line indices if requested
    if index:
        str_lines = lines
        lines = [index_key[line[6:11]] for line in lines]

    # Otherwise make new indices
    else:
        lines = [line[:6] + f"{i + 1:5d}" + line[11:] for i, line in enumerate(lines)]

    if kwargs.get('return_bonds', False):
        bonds = {tuple(sorted((idxmap[a], idxmap[b]))) for a, b in input_bonds}
        return lines, bonds

    return lines


def make_mda_uni(anames: ArrayLike,
                 atypes: ArrayLike,
                 resnames: ArrayLike,
                 resindices: ArrayLike,
                 resnums: ArrayLike,
                 segindices: ArrayLike,
                 segids: ArrayLike = None,
                 ) -> MDAnalysis.Universe:
    """
    Create an MDAnalysis universe from numpy arrays of atom information.

    Parameters
    ----------
    anames : ArrayLike
        Array of atom names. Length should be equal to the number of atoms.
    atypes : ArrayLike
        Array of atom elements or types. Length should be equal to the number of atoms.
    resnames : ArrayLike
        Array of residue names. Length should be equal to the number of residues.
    resindices : ArrayLike
        Array of residue indices. Length should be equal to the number of atoms. Elements of resindices should
        map to resnames and resnums of the atoms they represent.
    resnums : ArrayLike
        Array of residue numbers. Length should be equal to the number of residues.
    segindices : ArrayLike
        Array of segment/chain indices. Length should be equal to the number of residues.
    segids : ArrayLike, None
        Array of segment/chain IDs. Length should be equal to the number of segs/chains.

    Returns
    -------
    mda_uni : MDAnalysis.Universe
        The Universe created by the function.
    """

    n_atoms = len(anames)
    n_residues = len(np.unique(resindices))

    if segids is None:
        segids = np.array(["ABCDEFGHIJKLMNOPQRSTUVWXYZ"[i] for i in range(len(np.unique(segindices)))])

    mda_uni = mda.Universe.empty(
        n_atoms,
        n_residues=n_residues,
        atom_resindex=resindices,
        residue_segindex=segindices,
        trajectory=True,
    )

    mda_uni.add_TopologyAttr("type", atypes)
    mda_uni.add_TopologyAttr("resnum", resnums)
    mda_uni.add_TopologyAttr("resids", resnums)
    mda_uni.add_TopologyAttr("resname", resnames)
    mda_uni.add_TopologyAttr("name", anames)
    mda_uni.add_TopologyAttr("altLocs", [""] * len(atypes))
    mda_uni.add_TopologyAttr("segid")

    for i, segid in enumerate(segids):
        if i == 0:
            i_segment = mda_uni.segments[0]
            i_segment.segid = segid
        else:
            i_segment = mda_uni.add_Segment(segid=str(segid))

        mask = np.argwhere(np.asarray(segindices) == i).flatten()
        mda_uni.residues[mask.tolist()].segments = i_segment

    mda_uni.add_TopologyAttr(Segids(np.array(segids)))
    mda_uni.add_TopologyAttr(Atomindices())
    mda_uni.add_TopologyAttr(Resindices())
    mda_uni.add_TopologyAttr(Segindices())

    return mda_uni


DATA_DIR = Path(__file__).parent.absolute() / "data/"
RL_DIR = Path(__file__).parent.absolute() / "data/rotamer_libraries/"

# Define rotamer dihedral angle atoms
with open(DATA_DIR / "dihedral_defs.toml", "r") as f:
    dihedral_defs = rtoml.load(f)

with open(RL_DIR / "RotlibIndexes.pkl", "rb") as f:
    rotlib_indexes = pickle.load(f)

with open(DATA_DIR / 'BondDefs.pkl', 'rb') as f:
    bond_hmax_dict = {key: (val + 0.4 if 'H' in key else val + 0.35) for key, val in pickle.load(f).items()}
    bond_hmax_dict = defaultdict(lambda: 0, bond_hmax_dict)


    def bond_hmax(a): return bond_hmax_dict.get(tuple(i for i in a), 0)


    bond_hmax = np.vectorize(bond_hmax, signature="(n)->()")

atom_order = {"N": 0, "CA": 1, "C": 2, "O": 3, 'H': 5}

nataa_codes = {'ALA': 'A', 'CYS': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F', 'GLY': 'G', 'HIS': 'H', 'ILE': 'I',
               'LYS': 'K', 'LEU': 'L', 'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R', 'SER': 'S',
               'THR': 'T', 'VAL': 'V', 'TRP': 'W', 'TYR': 'Y'}

inataa = {val: key for key, val in nataa_codes.items()}

nataa_codes.update(inataa)
del inataa



import hashlib, os, pickle
from functools import partial
import numpy as np
import chilife
import pytest
import MDAnalysis as mda

ubq = mda.Universe("test_data/1ubq.pdb", in_memory=True)
U = mda.Universe("test_data/1omp.pdb", in_memory=True)
exou = mda.Universe("test_data/3tu3.pdb", in_memory=True)
traj = mda.Universe('test_data/xlsavetraj.pdb', in_memory=True)


def test_from_mda():
    res16 = U.residues[16]
    ensemble = chilife.RotamerEnsemble.from_mda(res16)
    ensemble.save_pdb("test_data/test_from_MDA.pdb")

    with open("test_data/ans_from_MDA.pdb", "r") as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("test_data/test_from_MDA.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove("test_data/test_from_MDA.pdb")
    assert ans == test


def test_with_sample():
    np.random.seed(200)

    SL = chilife.SpinLabel("R1C", 28, ubq, sample=2000)

    with np.load('test_data/withsample.npz') as f:
        ans = {key: f[key] for key in f}

    np.testing.assert_almost_equal(SL._coords, ans['coords'])
    np.testing.assert_almost_equal(SL.weights, ans['weights'])
    np.testing.assert_almost_equal(SL.dihedrals, ans['dihedrals'])

    assert len(SL._coords) == len(SL.internal_coords)


def test_user_label():
    SL = chilife.SpinLabel("TRT", 28, ubq, "A")
    chilife.save("28TRT.pdb", SL, KDE=False)

    with open("test_data/28TRT.pdb", "r") as f:
        nohstring1 = "".join((line for line in f.readlines() if line[76:79].strip() != 'H'))
        ans = hashlib.md5(nohstring1.encode('utf-8')).hexdigest()

    with open("28TRT.pdb", "r") as f:
        nohstring2 = "".join((line for line in f.readlines() if line[76:79].strip() != 'H'))
        test = hashlib.md5(nohstring2.encode('utf-8')).hexdigest()

    os.remove("28TRT.pdb")

    assert test == ans


def test_save_pkl():
    Lys = chilife.RotamerEnsemble("LYS")

    with open("test_data/Lys.pkl", "wb") as f:
        pickle.dump(Lys, f)

    with open("test_data/Lys.pkl", "rb") as f:
        reload = pickle.load(f)

    os.remove("test_data/Lys.pkl")

    assert Lys == reload


def test_save_pkl_2():
    res16 = U.residues[16]
    ensemble = chilife.RotamerEnsemble.from_mda(res16)
    with open("test_data/res16.pkl", "wb") as f:
        pickle.dump(ensemble, f)

    with open("test_data/res16.pkl", "rb") as f:
        reload = pickle.load(f)

    assert ensemble == reload
    os.remove("test_data/res16.pkl")


def test_sample():
    np.random.seed(200)
    K48 = chilife.RotamerEnsemble.from_mda(ubq.residues[47])
    coords, weight = K48.sample(off_rotamer=True)

    wans = 0.0037019925960148086
    cans = np.array([[20.76406623, 27.90259642, 22.59445098],
                     [21.54999924, 26.79599953, 23.13299942],
                     [23.02842999, 27.03777002, 22.93937713],
                     [23.39710135, 27.56595972, 21.90105423],
                     [21.1101072, 25.47090301, 22.45317533],
                     [21.19895831, 25.51252905, 20.90717057],
                     [21.03430238, 24.15607743, 20.2156025],
                     [22.11572213, 23.19412351, 20.72492909],
                     [22.16827211, 22.00913727, 19.85099131]])

    np.testing.assert_almost_equal(coords, cans)
    assert weight == wans


def test_multisample():
    R1M = chilife.SpinLabel.from_wizard("R1M", site=48, protein=ubq, to_find=10000)


@pytest.mark.parametrize("res", chilife.SUPPORTED_RESIDUES)
def test_lib_distribution_persists(res):
    if res in chilife.USER_dLIBRARIES:
        return None
    elif res in chilife.USER_LIBRARIES:
        L1 = chilife.SpinLabel(res)
        L2 = chilife.SpinLabel(res, sample=100)
    else:
        L1 = chilife.RotamerEnsemble(res)
        L2 = chilife.RotamerEnsemble(res, sample=100)

    np.testing.assert_almost_equal(L1._rdihedrals, L2._rdihedrals)
    np.testing.assert_almost_equal(L1._rsigmas, L2._rsigmas)
    np.testing.assert_almost_equal(L1._weights, L2._weights)


methods = ["rosetta", "bisect", "mmm", "fit"]


@pytest.mark.parametrize(("method"), methods)
def test_alignment_method(method):
    if method == "fit":
        with pytest.raises(NotImplementedError) as e_info:
            SL = chilife.SpinLabel(
                "R1C", site=28, protein=ubq, alignment_method=method, eval_clash=False
            )
            chilife.save(SL, ubq)
    else:
        SL = chilife.SpinLabel(
            "R1C",
            site=28,
            protein=ubq,
            alignment_method=method,
            energy_func=partial(chilife.get_lj_rep, forgive=0.8),
        )
        chilife.save(
            f"A28R1_{method}_aln_method.pdb", SL, KDE=False
        )

        with open(f"A28R1_{method}_aln_method.pdb", "r") as f:
            test = hashlib.md5(f.read().encode('utf-8')).hexdigest()

        with open(f"test_data/A28R1_{method}_aln_method.pdb", "r") as f:
            ans = hashlib.md5(f.read().encode('utf-8')).hexdigest()

        os.remove(f"A28R1_{method}_aln_method.pdb")
        assert ans == test


def test_catch_unused_kwargs():
    with pytest.raises(TypeError) as e_info:
        SL = chilife.SpinLabel("R1C", site=28, protein=ubq, fake_keyword="mmm")
    assert (
            str(e_info.value) == "Got unexpected keyword argument(s): fake_keyword"
    )


def test_guess_chain():
    anf = mda.Universe("test_data/1anf.pdb", in_memory=True)
    SL = chilife.SpinLabel.from_mmm("R1M", 20, forgive=0.9)


@pytest.mark.parametrize(("resi", "ans"), ((20, "A"), (206, "B")))
def test_guess_chain2(resi, ans):
    SL = chilife.SpinLabel("R1C", resi, exou)
    assert SL.chain == ans


@pytest.mark.parametrize("resi", (100, 344))
def test_guess_chain_fail(resi):
    with pytest.raises(ValueError) as e_info:
        SL = chilife.SpinLabel("R1C", resi, exou)


def test_mem_sample():
    SL1 = chilife.SpinLabel('R1M', 28, ubq, sample=10000)
    print('booger')


def test_label_as_library():
    R1C = chilife.RotamerEnsemble("R1C", site=28, protein=ubq)
    for ic in R1C.internal_coords:
        np.testing.assert_almost_equal(
            ic.coords[2], [38.73227962, 26.58109478, 12.6243569]
        )


def test_coord_setter0():
    R1C1 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq)


def test_coord_setter():
    R1C1 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq)
    R1C2 = chilife.RotamerEnsemble("R1C", site=29, protein=ubq)

    R1C1.coords = R1C2.coords

    np.testing.assert_allclose(R1C1.coords, R1C2.coords)
    np.testing.assert_allclose(R1C1.dihedrals, R1C2.dihedrals, rtol=1e-6)

    for key in 'ori', 'mx':
        np.testing.assert_allclose(R1C1.internal_coords[0].chain_operators[1][key],
                                   R1C2.internal_coords[0].chain_operators[1][key])


def test_coord_setter2():
    R1C1 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, sample=100)
    R1C2 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, sample=100)
    R1C1.coords = R1C2.coords[:, R1C2.side_chain_idx]

    np.testing.assert_allclose(R1C1.coords, R1C2.coords)
    np.testing.assert_allclose(R1C1.dihedrals, R1C2.dihedrals)


def test_dihedral_setter():
    R1C1 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, sample=100)
    R1C2 = chilife.RotamerEnsemble("R1C", site=28, protein=ubq, sample=100)
    R1C1.dihedrals = R1C2.dihedrals

    np.testing.assert_allclose(R1C1.coords, R1C2.coords)
    np.testing.assert_allclose(R1C1.dihedrals, R1C2.dihedrals)


def test_get_sasa():
    R1C = chilife.RotamerEnsemble("R1C")
    sasas = R1C.get_sasa()
    sasans = np.load('test_data/sasas.npy')
    np.testing.assert_allclose(sasas, sasans)


def test_default_dihedral_sigmas():
    rotlib = chilife.read_bbdep('R1C', -60, -50)
    SL = chilife.SpinLabel('R1C')
    np.testing.assert_allclose(rotlib['sigmas'][rotlib['sigmas'] != 0], SL.sigmas[SL.sigmas != 35.])


def test_construct_with_dihedral_sigmas():
    SL = chilife.SpinLabel('R1C', dihedral_sigmas=25)
    assert np.all(SL.sigmas == 25.)
    assert SL.sigmas.shape == (len(SL), len(SL.dihedral_atoms))


def test_construct_with_array_of_dihedral_sigmas():
    set_sigmas = [10, 20, 30, 40, 50]
    SL = chilife.SpinLabel('R1C', dihedral_sigmas=set_sigmas)
    for i in range(5):
        assert np.all(SL.sigmas[:, i] == set_sigmas[i])

    assert SL.sigmas.shape == (len(SL), len(SL.dihedral_atoms))


def test_construct_with_full_array_of_dihedral_sigmas():
    set_sigmas = np.random.rand(148, 5)
    SL = chilife.SpinLabel('R1C', dihedral_sigmas=set_sigmas)
    np.testing.assert_allclose(set_sigmas, SL.sigmas)


def test_dihedral_sigmas_fail():
    with pytest.raises(ValueError):
        SL = chilife.SpinLabel('R1C', dihedral_sigmas=[5, 2, 1])


def test_set_dihedral_sigmas():
    SL = chilife.SpinLabel('R1M')
    assert np.all(SL.sigmas == 35.)
    SL.set_dihedral_sampling_sigmas(25)
    assert np.all(SL.sigmas == 25.)


def test_clash_radius():
    RL = chilife.SpinLabel('TRT')
    assert RL.clash_radius == np.linalg.norm(RL.coords - RL.clash_ori, axis=-1).max() + 5


def test_nataa():
    SL = chilife.SpinLabel('R1M', 28, ubq)
    assert SL.nataa == 'A'


def test_from_trajectory():
    RE1 = chilife.RotamerEnsemble.from_trajectory(traj, 232, burn_in=0)
    RE2 = chilife.RotamerEnsemble.from_trajectory(traj, 231, burn_in=0)
    RE3 = chilife.RotamerEnsemble.from_trajectory(traj, 230, burn_in=0)

    assert len(RE1) == 1
    assert len(RE2) == 1
    assert len(RE3) == 10

    assert RE1.res == 'TRP'
    assert RE2.res == 'ALA'
    assert RE3.res == 'TRP'

    np.testing.assert_almost_equal(RE1.dihedrals, [[-72.49934635, 164.56177856]], decimal=5)
    np.testing.assert_almost_equal(RE2.dihedrals, [[]])
    np.testing.assert_almost_equal(RE3.dihedrals, [[-176.38805017,  -20.15226419],
                                                   [-176.38805017,  -52.83812431],
                                                   [-176.38805017, -111.39416684],
                                                   [-176.38805017,   73.81697923],
                                                   [-176.38805017, -134.54899132],
                                                   [-176.38805017,  118.2802729 ],
                                                   [-176.38805017,  164.59451953],
                                                   [  62.16344279,  -93.79344738],
                                                   [ -69.90840475,  -39.25944367],
                                                   [ -69.90840475,  161.67407146]], decimal=5)

    with pytest.raises(ValueError):
        RE = chilife.RotamerEnsemble.from_trajectory(traj, 232)

    SL1 = chilife.SpinLabel.from_trajectory(traj, 238, burn_in=0, spin_atoms=['N1', 'O1'])
    assert np.all(SL1.spin_atoms == np.array(['N1', 'O1']))

def test_to_rotlib():

    # Generate rotamer library from trajectory
    RE = chilife.RotamerEnsemble.from_trajectory(traj, 230, burn_in=0)
    RE.to_rotlib('Test')
    
    # Build rotamer ensemble from new rotamer library
    RE2 = chilife.RotamerEnsemble('TRP', rotlib='Test')

    # Load new rotamer libary, then delete file
    with np.load(f"Test_rotlib.npz", allow_pickle=True) as f:
        rotlib_test = dict(f)

    os.remove('Test_rotlib.npz')

    # Load previously saved reference rotamer libary
    with np.load(f"test_data/Test_rotlib.npz", allow_pickle=True) as f:
        rotlib_reference = dict(f)

    assert RE2.res == 'TRP'
    assert len(RE2) == 10

    np.testing.assert_almost_equal(RE2.dihedrals, [[-176.38805017,  -20.15226419],
                                                   [-176.38805017,  -52.83812431],
                                                   [-176.38805017, -111.39416684],
                                                   [-176.38805017,   73.81697923],
                                                   [-176.38805017, -134.54899132],
                                                   [-176.38805017,  118.2802729 ],
                                                   [-176.38805017,  164.59451953],
                                                   [  62.16344279,  -93.79344738],
                                                   [ -69.90840475,  -39.25944367],
                                                   [ -69.90840475,  161.67407146]], decimal=5)

    np.testing.assert_almost_equal(rotlib_test['coords'], rotlib_reference['coords'], decimal=5)
    np.testing.assert_almost_equal(rotlib_test['weights'], rotlib_reference['weights'], decimal=5)
    np.testing.assert_almost_equal(rotlib_test['dihedrals'], rotlib_reference['dihedrals'], decimal=5)


def test_sample_persists():
    rot1 = chilife.RotamerEnsemble('ARG', 28, ubq)
    rot2 = chilife.RotamerEnsemble('ARG', 28, ubq, eval_clash=True)

    np.random.seed(10)
    x = rot1.sample()

    np.random.seed(10)
    y = rot2.sample()

    np.testing.assert_allclose(x[0], y[0])
    assert x[1] == y[1]


def test_trim_false():
    rot1 = chilife.RotamerEnsemble('ARG', 28, ubq, eval_clash=True)
    rot2 = chilife.RotamerEnsemble('ARG', 28, ubq, eval_clash=True, trim=False)
    rot3 = chilife.RotamerEnsemble('ARG', 28, ubq)

    assert len(rot2) == len(rot3)
    most_probable = np.sort(rot2.weights)[::-1][:len(rot1)]
    most_probable /= most_probable.sum()
    np.testing.assert_almost_equal(rot1.weights, most_probable)
    assert np.any(np.not_equal(rot2.weights, rot3.weights))

def test_min_callback():
    vals = []
    ivals = []
    def my_callback(val, i):
        vals.append(val)
        ivals.append(i)

    SL1 = chilife.RotamerEnsemble('ARG', 28, ubq)
    SL1.minimize(callback=my_callback)

    assert len(vals) > 0
    assert len(ivals) > 0


def test_copy_custom_lib():
    XYZ41 = chilife.SpinLabel('XYZ', 28, ubq, rotlib='test_data/usr_rtlb/XYZ_rotlib.npz')
    x2 = XYZ41.copy()


def test_from_wizard_custom_rotlib():
    np.random.seed(0)
    A28NBA = chilife.SpinLabel.from_wizard('TES', 28, ubq, rotlib='test_data/usr_rtlb/NBA_rotlib.npz')
    ans = np.load('test_data/from_wiz_cust.npy')
    np.testing.assert_almost_equal(A28NBA.coords, ans, decimal=5)


